"""
Module to provide processing for the container blocks.
"""

from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING, List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_block_leaf_processor import (
    ContainerBlockLeafProcessor,
)
from pymarkdown.container_blocks.container_block_non_leaf_processor import (
    ContainerBlockNonLeafProcessor,
)
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.extensions.pragma_token import PragmaExtension
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken

if TYPE_CHECKING:  # pragma: no cover
    from pymarkdown.container_blocks.parse_block_pass_properties import (
        ParseBlockPassProperties,
    )

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class ContainerBlockProcessor:
    """
    Class to provide processing for the container blocks.
    """

    @staticmethod
    def __setup(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[PositionMarker, bool, bool]:
        POGGER.debug(">>")
        POGGER.debug(">>")
        if ContainerBlockProcessor.__look_for_pragmas(
            position_marker,
            grab_bag,
        ):
            return position_marker, True, False

        position_marker = ContainerBlockProcessor.__prepare_container_start_variables(
            parser_state,
            position_marker,
            grab_bag,
        )

        ContainerBlockProcessor.__prepare_container_start_variables2(
            parser_state, position_marker, grab_bag
        )

        is_not_in_root_list = not (
            parser_state.token_stack
            and len(parser_state.token_stack) >= 2
            and parser_state.token_stack[1].is_list
        )

        POGGER.debug(
            "position_marker>:$:$:",
            position_marker.index_number,
            position_marker.text_to_parse,
        )
        return (
            position_marker,
            False,
            is_not_in_root_list,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def parse_line_for_container_blocks(
        parser_state: ParserState,
        position_marker: PositionMarker,
        ignore_link_definition_start: bool,
        ignore_table_start: bool,
        parser_properties: ParseBlockPassProperties,
        container_start_bq_count: int,
        container_depth: int = 0,
        adjusted_block_index: Optional[int] = None,
        initial_block_quote_count: Optional[int] = None,
        original_line: Optional[str] = None,
    ) -> Tuple[
        List[MarkdownToken],
        Optional[str],
        Optional[int],
        Optional[RequeueLineInfo],
        bool,
        bool,
    ]:
        """
        Parse the line, taking care to handle any container blocks before deciding
        whether or not to pass the (remaining parts of the) line to the leaf block
        processor.

        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """

        grab_bag = ContainerGrabBag(
            parser_state,
            container_depth,
            initial_block_quote_count,
            adjusted_block_index,
            container_start_bq_count,
            parser_properties,
            ignore_link_definition_start,
            ignore_table_start,
            (
                original_line
                if original_line is not None
                else position_marker.text_to_parse
            ),
            ContainerBlockProcessor.parse_line_for_container_blocks,
        )
        (
            position_marker,
            did_find_pragma,
            is_not_in_root_list,
        ) = ContainerBlockProcessor.__setup(
            parser_state,
            position_marker,
            grab_bag,
        )
        if did_find_pragma:
            return [], None, None, None, False, False
        # POGGER.debug("position_marker.index_number>>$", position_marker.index_number)

        # POGGER.debug(">>parser_state.token_stack:$", parser_state.token_stack)
        # POGGER.debug(">>is_not_in_root_list=:$:", is_not_in_root_list)
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Must have whitespace by this point."
        grab_bag.bogus = None
        if (
            not grab_bag.container_depth
            and len(grab_bag.extracted_whitespace) >= 4
            and is_not_in_root_list
        ):
            POGGER.debug("indent")
            ContainerBlockProcessor.__handle_indented_block_start(
                parser_state, position_marker, grab_bag
            )
        else:
            ContainerBlockNonLeafProcessor.handle_non_leaf_block(
                parser_state,
                position_marker,
                grab_bag,
            )

        if (
            grab_bag.can_continue
            and not grab_bag.do_skip_containers_before_leaf_blocks
            and not grab_bag.did_blank
        ) or grab_bag.do_force_leaf_token_parse:
            ContainerBlockLeafProcessor.handle_leaf_tokens(
                parser_state,
                position_marker,
                grab_bag,
            )

        return (
            grab_bag.container_tokens,
            grab_bag.line_to_parse,
            grab_bag.block_quote_data.current_count,
            grab_bag.requeue_line_info,
            grab_bag.did_blank,
            grab_bag.do_force_list_continuation,
        )
        # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_indented_block_start(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        (
            grab_bag.can_continue,
            grab_bag.line_to_parse,
            grab_bag.start_index,
            grab_bag.text_removed_by_container,
            grab_bag.removed_chars_at_start_of_line,
        ) = (
            True,
            position_marker.text_to_parse,
            position_marker.index_number,
            "",
            0,
        )
        POGGER.debug("parser_state.token_stack>>$", parser_state.token_stack)
        is_paragraph_continuation = (
            parser_state.token_stack and parser_state.token_stack[-1].is_paragraph
        )
        list_index = parser_state.find_last_list_block_on_stack()
        block_index = parser_state.find_last_block_quote_on_stack()
        POGGER.debug("list_index>>$", list_index)
        POGGER.debug("block_index>>$", block_index)
        if is_paragraph_continuation and block_index > list_index:
            grab_bag.was_paragraph_continuation = True
        # if is_paragraph_continuation and block_index < list_index:
        #     grab_bag.was_other_paragraph_continuation = True
        if (
            not is_paragraph_continuation
            and parser_state.token_stack
            and len(parser_state.token_stack) >= 2
            and parser_state.token_stack[1].is_block_quote
        ):
            x_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                include_lists=True,
                include_block_quotes=True,
            )
            # POGGER.debug("x_tokens=:$:", x_tokens)
            grab_bag.extend_container_tokens(x_tokens)

    @staticmethod
    def __prepare_container_start_variables(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> PositionMarker:
        # POGGER.debug("Stack Depth:$:", parser_state.original_stack_depth)
        # POGGER.debug("Document Depth:$:", parser_state.original_document_depth)
        if grab_bag.container_depth:
            return position_marker

        if ParserHelper.tab_character in position_marker.text_to_parse:
            detabified_line = TabHelper.detabify_string(position_marker.text_to_parse)
            POGGER.debug("Before tab replacement:$:", position_marker.text_to_parse)
            POGGER.debug("After tab replacement :$:", detabified_line)
            position_marker = PositionMarker(
                position_marker.line_number,
                position_marker.index_number,
                detabified_line,
                position_marker.index_indent,
            )

        parser_state.mark_start_information(position_marker)

        parser_state.copy_of_token_stack = []
        parser_state.copy_of_token_stack.extend(parser_state.token_stack)

        parser_state.block_copy = [
            copy.deepcopy(i.matching_markdown_token)
            for i in parser_state.token_stack
            if not i.is_document
        ]

        return position_marker

    @staticmethod
    def __prepare_container_start_variables2(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        # Debug to be used for block quotes if needed.
        # POGGER.debug(
        #    "Last Block Quote:$:",
        #    parser_state.last_block_quote_stack_token,
        # )
        # POGGER.debug(
        #    "Last Block Quote:$:",
        #    parser_state.last_block_quote_markdown_token_index,
        # )
        # POGGER.debug(
        #    "Last Block Quote:$:", parser_state.copy_of_last_block_quote_markdown_token
        # )

        (
            new_start_index,
            grab_bag.extracted_whitespace,
        ) = ParserHelper.extract_spaces_verified(position_marker.text_to_parse, 0)
        grab_bag.start_index = new_start_index
        ContainerBlockProcessor.__calculate_for_container_blocks(
            parser_state,
            grab_bag,
        )

    @staticmethod
    def __calculate_for_container_blocks(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
    ) -> None:
        """
        Perform some calculations that will be needed for parsing the container blocks.
        """
        grab_bag.current_container_blocks = [
            ind for ind in parser_state.token_stack if ind.is_list
        ]
        grab_bag.block_quote_data = BlockQuoteData(
            (
                0
                if grab_bag.initial_block_quote_count is None
                else grab_bag.initial_block_quote_count
            ),
            parser_state.count_of_block_quotes_on_stack(),
        )

        ContainerBlockProcessor.__calculate_adjusted_whitespace(
            parser_state,
            grab_bag,
        )

    @staticmethod
    def __look_back_in_document_for_block_quote(
        parser_state: ParserState, token_index: int
    ) -> Optional[BlockQuoteMarkdownToken]:
        # Option? Look back on stack instead?
        other_block_quote_token, other_token_index = None, token_index
        while other_token_index >= 0:
            if parser_state.token_document[other_token_index].is_block_quote_start:
                block_quote_token = cast(
                    BlockQuoteMarkdownToken,
                    parser_state.token_document[other_token_index],
                )
                if not ContainerBlockProcessor.__does_block_quote_token_have_end(
                    parser_state, block_quote_token
                ):
                    other_block_quote_token = block_quote_token
                    break
            other_token_index -= 1
        POGGER.debug_with_visible_whitespace(
            "PLFCB>>other_block_quote_token>>$",
            other_block_quote_token,
        )
        return other_block_quote_token

    @staticmethod
    def __calculate_adjusted_whitespace_kludge_without_found(
        parser_state: ParserState, token_index: int, grab_bag: ContainerGrabBag
    ) -> Tuple[bool, int]:
        other_block_quote_token = (
            ContainerBlockProcessor.__look_back_in_document_for_block_quote(
                parser_state, token_index
            )
        )

        # Check to see if out first block token is the same as our first.
        # if not, do not use it as a base.
        #
        # Note; may need to be tweaked for extra levels.
        if not grab_bag.container_depth and other_block_quote_token:
            POGGER.debug(
                "parser_state.token_stack[1]>>:$:", parser_state.token_stack[1]
            )
            if (
                parser_state.token_stack[1].matching_markdown_token
                != other_block_quote_token
            ):
                other_block_quote_token = None

        if other_block_quote_token:
            POGGER.debug("PLFCB>>other_block_quote_token>>:$:", other_block_quote_token)
            POGGER.debug(
                "PLFCB>>other_block_quote_token.leading_text_index>>:$:",
                other_block_quote_token.leading_text_index,
            )
            leading_spaces = other_block_quote_token.calculate_next_bleading_space_part(
                increment_index=False, delta=-1
            )
            POGGER.debug("PLFCB>>leading_spaces>>:$:", leading_spaces)
            POGGER.debug("PLFCB>>other_block_quote_token>>:$:", other_block_quote_token)
            POGGER.debug(
                "PLFCB>>other_block_quote_token.leading_text_index>>:$:",
                other_block_quote_token.leading_text_index,
            )
            force_reline = True
            old_start_index = len(leading_spaces)
        else:
            force_reline = False
            list_token = cast(
                ListStartMarkdownToken, parser_state.token_document[token_index]
            )
            old_start_index = list_token.indent_level
        return force_reline, old_start_index

    @staticmethod
    def __does_block_quote_token_have_end(
        parser_state: ParserState,
        found_block_quote_token: Optional[BlockQuoteMarkdownToken],
    ) -> bool:
        assert (
            found_block_quote_token
        ), "If we are processing a block quote, need a token."
        POGGER.debug(
            "PLFCB>>found_block_quote_token>>:$:",
            ParserHelper.make_value_visible(found_block_quote_token),
        )
        found_token_index = parser_state.token_document.index(found_block_quote_token)
        POGGER.debug("PLFCB>>found_token_index>>:$:", found_token_index)
        found_token_index += 1
        while found_token_index < len(parser_state.token_document):
            if parser_state.token_document[found_token_index].is_end_token:
                end_token = cast(
                    EndMarkdownToken, parser_state.token_document[found_token_index]
                )
                if end_token.start_markdown_token == found_block_quote_token:
                    break
            found_token_index += 1
        POGGER.debug(
            "PLFCB>>found_token_index>>:$:max:$",
            found_token_index,
            len(parser_state.token_document),
        )
        return found_token_index < len(parser_state.token_document)

    @staticmethod
    def __calculate_adjusted_whitespace_kludge(
        parser_state: ParserState,
        token_index: int,
        found_block_quote_token: Optional[BlockQuoteMarkdownToken],
        grab_bag: ContainerGrabBag,
    ) -> None:
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Must have whitespace by this point."
        previous_ws_len = 0
        force_reline, ws_len = (
            False,
            TabHelper.calculate_length(grab_bag.extracted_whitespace) + previous_ws_len,
        )
        if (
            found_block_quote_token
            and ContainerBlockProcessor.__does_block_quote_token_have_end(
                parser_state, found_block_quote_token
            )
        ):
            found_block_quote_token = None
        if found_block_quote_token:
            POGGER.debug(
                "PLFCB>>found_block_quote_token>>:$:",
                ParserHelper.make_value_visible(found_block_quote_token),
            )
            leading_spaces = found_block_quote_token.calculate_next_bleading_space_part(
                increment_index=False, delta=-1, allow_overflow=True
            )
            POGGER.debug("PLFCB>>leading_spaces>>:$:", leading_spaces)
            old_start_index = len(leading_spaces)
        else:
            (
                force_reline,
                old_start_index,
            ) = ContainerBlockProcessor.__calculate_adjusted_whitespace_kludge_without_found(
                parser_state, token_index, grab_bag
            )
        POGGER.debug(
            "old_start_index>>$>>ws_len>>$>>force_reline>>$",
            old_start_index,
            ws_len,
            force_reline,
        )
        if force_reline or ws_len >= old_start_index:
            POGGER.debug("RELINE:$:", grab_bag.line_to_parse)
            grab_bag.adj_ws = grab_bag.extracted_whitespace[old_start_index:]
            POGGER.debug(
                "adj_ws:$: old_start_index:$:", grab_bag.adj_ws, old_start_index
            )

    @staticmethod
    def __look_for_any_list_start(
        parser_state: ParserState,
    ) -> Tuple[int, Optional[BlockQuoteMarkdownToken]]:
        found_block_quote_token: Optional[BlockQuoteMarkdownToken] = None
        token_index = len(parser_state.token_document) - 1
        while token_index >= 0 and not (
            parser_state.token_document[token_index].is_any_list_token
        ):
            if (
                not found_block_quote_token
                and parser_state.token_document[token_index].is_block_quote_start
            ):
                found_block_quote_token = cast(
                    BlockQuoteMarkdownToken,
                    parser_state.token_document[token_index],
                )
            token_index -= 1
        POGGER.debug(
            "PLFCB>>Started list-last token>>$",
            parser_state.token_document[token_index],
        )

        POGGER.debug(
            "CAW>>found_block_quote_token>>:$:",
            ParserHelper.make_value_visible(found_block_quote_token),
        )
        if found_block_quote_token:
            POGGER.debug(
                "PLFCB>>leading_text_index>>$",
                found_block_quote_token.leading_text_index,
            )
        assert token_index >= 0, "Token index should be positive by this point."
        return token_index, found_block_quote_token

    @staticmethod
    def __calculate_adjusted_whitespace(
        parser_state: ParserState,
        grab_bag: ContainerGrabBag,
    ) -> None:
        """
        Based on the last container on the stack, determine what the adjusted whitespace is.
        """

        grab_bag.adj_ws = grab_bag.extracted_whitespace
        # assert (
        #     grab_bag.adj_ws is not None
        # ), "Adjusted whitespace must be set by this point."

        last_block_stack_index = parser_state.find_last_list_block_on_stack()
        if last_block_stack_index <= 0:
            assert (
                not grab_bag.current_container_blocks
            ), "Must not have container blocks by this point."
            POGGER.debug("PLFCB>>No Started lists")
            if grab_bag.adjusted_block_index is None:
                POGGER.debug("PLFCB>>No Started Block Quote")
            else:
                POGGER.debug("PLFCB>>Started Block Quote")
                assert (
                    grab_bag.extracted_whitespace is not None
                ), "Must have whitespace by this point."
                grab_bag.adj_ws = grab_bag.extracted_whitespace[
                    grab_bag.adjusted_block_index :
                ]
        else:
            assert (
                grab_bag.current_container_blocks is not None
            ), "Must have container blocks by this point."
            POGGER.debug(
                "PLFCB>>Started list-last stack>>$",
                parser_state.token_stack,
            )
            POGGER.debug(
                "PLFCB>>Started list-last stack>>$",
                parser_state.token_stack[last_block_stack_index],
            )

            (
                token_index,
                found_block_quote_token,
            ) = ContainerBlockProcessor.__look_for_any_list_start(parser_state)

            # assert grab_bag.adj_ws is not None, "TOoDO: why here?"
            ContainerBlockProcessor.__calculate_adjusted_whitespace_kludge(
                parser_state,
                token_index,
                found_block_quote_token,
                grab_bag,
            )
            # assert grab_bag.adj_ws is not None, "Adjusted whitespace must be set by this point."

    @staticmethod
    def __look_for_pragmas(
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        _, pragma_whitespace = ParserHelper.extract_spaces(
            position_marker.text_to_parse, 0
        )
        return (
            PragmaExtension.look_for_pragmas(
                position_marker,
                position_marker.text_to_parse,
                grab_bag.container_depth,
                pragma_whitespace,
                grab_bag.parser_properties,
            )
            if grab_bag.parser_properties.is_pragmas_enabled
            else False
        )


# pylint: enable=too-few-public-methods



"""
Module to provide transformations for a list block.
"""

import copy
import logging
from typing import List, Optional, Tuple, Union, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.transform_markdown.markdown_transform_context import (
    IndentAdjustment,
    MarkdownTransformContext,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class TransformListBlock:
    """
    Class to provide transformations for a list block.
    """

    @staticmethod
    def rehydrate_list_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
        transformed_data: str,
    ) -> str:
        """
        Rehydrate the unordered list start token.
        """

        assert (
            next_token is not None
        ), "If there is a start token, must be at least an end token."
        POGGER.debug(
            f">>current_token>>{ParserHelper.make_value_visible(current_token)}<<"
        )
        current_list_token = cast(ListStartMarkdownToken, current_token)

        extracted_whitespace = current_list_token.extracted_whitespace
        POGGER.debug(f">>extracted_whitespace>>{extracted_whitespace}<<")
        had_weird_block_quote_in_list = False
        if previous_token:
            (
                previous_indent,
                extracted_whitespace,
                was_within_block_token,
                post_adjust_whitespace,
                _,
                had_weird_block_quote_in_list,
            ) = TransformListBlock.rehydrate_list_start_previous_token(
                context,
                current_list_token,
                previous_token,
                extracted_whitespace,
            )
            POGGER.debug(f">>extracted_whitespace>>{extracted_whitespace}<<")
            POGGER.debug(f">>post_adjust_whitespace>>{post_adjust_whitespace}<<")
        else:
            previous_indent, post_adjust_whitespace, was_within_block_token = (
                0,
                None,
                False,
            )

        POGGER.debug(
            f">>had_weird_block_quote_in_list>>{had_weird_block_quote_in_list}<<"
        )
        context.container_token_stack.append(copy.deepcopy(current_list_token))
        context.original_container_token_stack.append(current_list_token)
        context.container_token_indents.append(IndentAdjustment())

        POGGER.debug(f">>extracted_whitespace>>{extracted_whitespace}<<")
        POGGER.debug(
            f">>transformed_data>>{ParserHelper.make_value_visible(transformed_data)}<<"
        )

        if was_within_block_token:
            adjustment_since_newline = 0
        else:
            (
                adjustment_since_newline,
                extracted_whitespace,
            ) = TransformListBlock.adjust_whitespace_for_block_quote(
                transformed_data, extracted_whitespace
            )
        POGGER.debug(f">>adjustment_since_newline>>{adjustment_since_newline}<<")
        POGGER.debug(f">>extracted_whitespace>>{extracted_whitespace}<<")

        return TransformListBlock.__rehydrate_list_start_calculate_start(
            current_list_token,
            next_token,
            extracted_whitespace,
            previous_indent,
            adjustment_since_newline,
            post_adjust_whitespace,
        )

    @staticmethod
    def rehydrate_list_start_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        actual_tokens: List[MarkdownToken],
        token_index: int,
    ) -> str:
        """
        Rehydrate the ordered list end token.
        """
        _ = actual_tokens, token_index
        del context.container_token_stack[-1]
        del context.original_container_token_stack[-1]
        del context.container_token_indents[-1]

        current_end_token = cast(EndMarkdownToken, current_token)
        current_start_token = cast(
            ListStartMarkdownToken, current_end_token.start_markdown_token
        )

        leading_spaces_index, expected_leading_spaces_index = (
            current_start_token.leading_spaces_index,
            ParserHelper.count_newlines_in_text(
                current_start_token.extracted_whitespace
            ),
        )

        assert leading_spaces_index == expected_leading_spaces_index, (
            f"leading_spaces_index={leading_spaces_index};"
            + f"expected_leading_spaces_index={expected_leading_spaces_index}"
        )
        return ""

    @staticmethod
    def adjust_whitespace_for_block_quote(
        transformed_data: str, extracted_whitespace: str
    ) -> Tuple[int, str]:
        """
        Make sure lists can adjust for block quote.
        """
        transformed_data_since_newline = transformed_data
        if ParserHelper.newline_character in transformed_data_since_newline:
            last_newline_index = transformed_data_since_newline.rindex(
                ParserHelper.newline_character
            )
            transformed_data_since_newline = transformed_data_since_newline[
                last_newline_index + 1 :
            ]
        adjustment_since_newline = 0
        # transformed_data_since_newline_size = len(
        #     transformed_data_since_newline
        # )
        POGGER.debug(
            f">>transformed_data_since_newline>>:{transformed_data_since_newline}:<<"
        )
        # POGGER.debug(f">>adjustment_since_newline>>:{adjustment_since_newline}:<<")
        # POGGER.debug(
        #     f">>transformed_data_since_newline_size>>:{transformed_data_since_newline_size}:<<"
        # )
        # POGGER.debug(f">>extracted_whitespace>>:{extracted_whitespace}:<<")
        # if (
        #     extracted_whitespace
        #     and len(extracted_whitespace) >= transformed_data_since_newline_size
        #     and ">" in transformed_data_since_newline
        # ):
        #     adjustment_since_newline = transformed_data_since_newline_size
        #     extracted_whitespace = extracted_whitespace[adjustment_since_newline:]
        POGGER.debug(f">>adjustment_since_newline>>:{adjustment_since_newline}:<<")
        POGGER.debug(f">>extracted_whitespace>>:{extracted_whitespace}:<<")
        return adjustment_since_newline, extracted_whitespace

    @staticmethod
    def rehydrate_list_start_previous_token(
        context: MarkdownTransformContext,
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        previous_token: MarkdownToken,
        extracted_whitespace: str,
    ) -> Tuple[int, str, bool, Optional[str], bool, bool]:
        """
        Rehydrate the list start previous.
        """
        (
            previous_indent,
            post_adjust_whitespace,
            was_within_block_token,
            containing_block_quote_token,
            containing_list_token,
            deeper_containing_block_quote_token,
        ) = TransformListBlock.__rehydrate_list_start_previous_token_start(
            context, current_token, previous_token, extracted_whitespace
        )

        had_weird_block_quote_in_list = False
        did_container_start_midline = False
        if previous_token.is_list_start:
            POGGER.debug("rlspt>>is_list_start")
            previous_list_token = cast(ListStartMarkdownToken, previous_token)
            (
                previous_indent,
                extracted_whitespace,
            ) = TransformListBlock.__rehydrate_list_start_prev_list(
                current_token, previous_list_token
            )
        elif previous_token.is_block_quote_start:
            POGGER.debug("rlspt>>is_block_quote_start")
            assert (
                containing_block_quote_token is not None
            ), "Indicated a nested block start, must have a token."
            (
                previous_indent,
                post_adjust_whitespace,
                extracted_whitespace,
            ) = TransformListBlock.__rehydrate_list_start_prev_block_quote(
                current_token,
                previous_token,
                containing_block_quote_token,
            )
        elif containing_block_quote_token:
            POGGER.debug("rlspt>>containing_block_quote_token")
            (
                was_within_block_token,
                previous_indent,
            ) = TransformListBlock.__rehydrate_list_start_contained_in_block_quote(
                current_token, containing_block_quote_token
            )
        elif containing_list_token:
            POGGER.debug("rlspt>>containing_list_token")
            (
                previous_indent,
                extracted_whitespace,
                post_adjust_whitespace,
                did_container_start_midline,
                had_weird_block_quote_in_list,
            ) = TransformListBlock.__rehydrate_list_start_contained_in_list(
                context,
                current_token,
                containing_list_token,
                deeper_containing_block_quote_token,
                extracted_whitespace,
                previous_token,
            )

        POGGER.debug(f"xx>>previous_indent:{previous_indent}:")
        POGGER.debug(f"xx>>extracted_whitespace:{extracted_whitespace}:")
        POGGER.debug(f"xx>>was_within_block_token:{was_within_block_token}:")
        POGGER.debug(f"xx>>post_adjust_whitespace:{post_adjust_whitespace}:")
        POGGER.debug(f"xx>>did_container_start_midline:{did_container_start_midline}:")
        return (
            previous_indent,
            extracted_whitespace,
            was_within_block_token,
            post_adjust_whitespace,
            did_container_start_midline,
            had_weird_block_quote_in_list,
        )

    @staticmethod
    def __rehydrate_list_start_previous_token_start(
        context: MarkdownTransformContext,
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        previous_token: MarkdownToken,
        extracted_whitespace: str,
    ) -> Tuple[
        int,
        Optional[str],
        bool,
        Optional[BlockQuoteMarkdownToken],
        Optional[ListStartMarkdownToken],
        Optional[BlockQuoteMarkdownToken],
    ]:
        previous_indent, was_within_block_token = 0, False
        post_adjust_whitespace: Optional[str] = None
        POGGER.debug(
            f"rlspt>>current_token>>{ParserHelper.make_value_visible(current_token)}<<"
        )
        POGGER.debug(
            f"rlspt>>previous_token>>{ParserHelper.make_value_visible(previous_token)}<<"
        )
        POGGER.debug(
            f"rlspt>>extracted_whitespace>>{ParserHelper.make_value_visible(extracted_whitespace)}<<"
        )
        POGGER.debug(
            f"rls>>self.context.container_token_stack>>{ParserHelper.make_value_visible(context.container_token_stack)}<<"
        )
        containing_block_quote_token = TransformListBlock.__look_for_last_block_token(
            context
        )
        POGGER.debug(
            f"rls>>containing_block_quote_token>>{ParserHelper.make_value_visible(containing_block_quote_token)}<<"
        )

        token_stack_index = len(context.container_token_stack) - 1
        POGGER.debug(f"rls>>token_stack_index2>>{token_stack_index}<<")

        containing_list_token, deeper_containing_block_quote_token = None, None
        if (
            token_stack_index >= 0
            and containing_block_quote_token
            != context.container_token_stack[token_stack_index]
        ):
            containing_list_token = cast(
                ListStartMarkdownToken, context.container_token_stack[token_stack_index]
            )
            deeper_containing_block_quote_token = containing_block_quote_token
            containing_block_quote_token = None

        return (
            previous_indent,
            post_adjust_whitespace,
            was_within_block_token,
            containing_block_quote_token,
            containing_list_token,
            deeper_containing_block_quote_token,
        )

    @staticmethod
    def __rehydrate_list_start_prev_list(
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        previous_token: ListStartMarkdownToken,
    ) -> Tuple[int, str]:
        _ = current_token
        return previous_token.indent_level, ""

    @staticmethod
    def __rehydrate_list_start_prev_block_quote(
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        previous_token: MarkdownToken,
        containing_block_quote_token: BlockQuoteMarkdownToken,
    ) -> Tuple[int, str, str]:
        previous_block_token = cast(BlockQuoteMarkdownToken, previous_token)
        assert (
            previous_block_token.bleading_spaces is not None
        ), "Bleading spaces must be defined by now."
        previous_indent = (
            len(
                previous_block_token.calculate_next_bleading_space_part(
                    increment_index=False
                )
            )
            if ParserHelper.newline_character in previous_block_token.bleading_spaces
            else len(previous_block_token.bleading_spaces)
        )
        POGGER.debug(
            f"adj->current_token>>:{ParserHelper.make_value_visible(current_token)}:<<"
        )
        POGGER.debug(
            f"adj->containing_block_quote_token>>:{ParserHelper.make_value_visible(containing_block_quote_token)}:<<"
        )
        assert (
            current_token.line_number == containing_block_quote_token.line_number
        ), "Line numbers must be the same."
        assert (
            containing_block_quote_token.bleading_spaces is not None
        ), "Bleading spaces must be defined by now."
        split_leading_spaces = containing_block_quote_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        block_quote_leading_space = split_leading_spaces[0]
        block_quote_leading_space_length = len(block_quote_leading_space)

        POGGER.debug(
            f"bq->len>>:{block_quote_leading_space}: {block_quote_leading_space_length}"
        )

        post_adjust_whitespace = "".ljust(
            current_token.column_number - block_quote_leading_space_length - 1, " "
        )
        extracted_whitespace = ""
        POGGER.debug(
            f"post_adjust_whitespace:{post_adjust_whitespace}: extracted_whitespace:{extracted_whitespace}:"
        )
        return previous_indent, post_adjust_whitespace, extracted_whitespace

    @staticmethod
    def __rehydrate_list_start_contained_in_block_quote(
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        containing_block_quote_token: BlockQuoteMarkdownToken,
    ) -> Tuple[bool, int]:
        block_quote_leading_space = (
            containing_block_quote_token.calculate_next_bleading_space_part(
                increment_index=False, delta=-1
            )
        )
        previous_indent = len(block_quote_leading_space)
        POGGER.debug(f"adj->rls>>previous_indent>>:{previous_indent}:<<")
        POGGER.debug(
            f"adj->rls>>current_token.indent_level>>:{current_token.indent_level}:<<"
        )

        return True, previous_indent

    # pylint: disable=too-many-arguments
    @staticmethod
    def __rehydrate_list_start_contained_in_list(
        context: MarkdownTransformContext,
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        containing_list_token: ListStartMarkdownToken,
        deeper_containing_block_quote_token: Optional[BlockQuoteMarkdownToken],
        extracted_whitespace: str,
        previous_token: MarkdownToken,
    ) -> Tuple[int, str, str, bool, bool]:
        # POGGER.debug(
        #     f"adj->containing_list_token>>:{ParserHelper.make_value_visible(containing_list_token)}:<<"
        # )
        # POGGER.debug(
        #     f"adj->deeper_containing_block_quote_token>>:{ParserHelper.make_value_visible(deeper_containing_block_quote_token)}:<<"
        # )

        # POGGER.debug(
        #     f"adj->extracted_whitespace>>:{ParserHelper.make_value_visible(extracted_whitespace)}:<<"
        # )
        (
            starting_whitespace,
            did_container_start_midline,
            block_quote_leading_space_length,
            had_weird_block_quote_in_list,
            list_leading_space_length,
        ) = TransformListBlock.__rehydrate_list_start_contained_in_list_start(
            context, previous_token, current_token, deeper_containing_block_quote_token
        )

        list_start_content_length = 0
        # POGGER.debug(
        #     f"previous_token-->{ParserHelper.make_value_visible(previous_token)}"
        # )
        # POGGER.debug(
        #     f"current_token-->{ParserHelper.make_value_visible(current_token)}"
        # )
        # POGGER.debug(f"next_token-->{ParserHelper.make_value_visible(next_token)}")
        if (
            current_token.is_new_list_item
            and previous_token.is_end_token
            and previous_token.is_block_quote_end
        ):
            new_list_token = cast(NewListItemMarkdownToken, containing_list_token)
            list_start_content_length = (
                len(new_list_token.list_start_content)
                if new_list_token.is_ordered_list_start
                else 0
            )

        post_adjust_whitespace = TransformListBlock.__calculate_post_adjust_whitespace(
            starting_whitespace,
            containing_list_token,
            block_quote_leading_space_length,
            list_leading_space_length,
            list_start_content_length,
            current_token,
        )

        (
            previous_indent,
            extracted_whitespace,
        ) = TransformListBlock.__rehydrate_list_start_contained_in_list_spacingx(
            containing_list_token, current_token, block_quote_leading_space_length
        )
        # POGGER.debug(f"adj->post_adjust_whitespace>>:{post_adjust_whitespace}:<<")
        return (
            previous_indent,
            extracted_whitespace,
            post_adjust_whitespace,
            did_container_start_midline,
            had_weird_block_quote_in_list,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __look_for_last_block_token(
        context: MarkdownTransformContext,
    ) -> Optional[BlockQuoteMarkdownToken]:
        # found_block_token: Optional[BlockQuoteMarkdownToken] = None
        found_token = next(
            (
                context.container_token_stack[i]
                for i in range(len(context.container_token_stack) - 1, -1, -1)
                if context.container_token_stack[i].is_block_quote_start
            ),
            None,
        )
        POGGER.debug(
            f">>found_block_token>>{ParserHelper.make_value_visible(found_token)}<"
        )
        return cast(BlockQuoteMarkdownToken, found_token) if found_token else None

    @staticmethod
    def __rehydrate_list_start_contained_in_list_start(
        context: MarkdownTransformContext,
        previous_token: MarkdownToken,
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
        deeper_containing_block_quote_token: Optional[BlockQuoteMarkdownToken],
    ) -> Tuple[str, bool, int, bool, int]:
        starting_whitespace = ""
        check_list_for_indent = True
        did_container_start_midline = False
        block_quote_leading_space_length = 0
        had_weird_block_quote_in_list = False
        list_leading_space_length = 0

        if deeper_containing_block_quote_token:
            (
                check_list_for_indent,
                starting_whitespace,
                did_container_start_midline,
                block_quote_leading_space_length,
                had_weird_block_quote_in_list,
            ) = TransformListBlock.__rehydrate_list_start_contained_in_list_deeper_block_quote(
                context,
                previous_token,
                deeper_containing_block_quote_token,
                current_token,
            )

        if (
            check_list_for_indent
            and previous_token
            and previous_token.line_number == current_token.line_number
            and previous_token.is_new_list_item
        ):
            previous_new_list_token = cast(NewListItemMarkdownToken, previous_token)
            list_leading_space_length = previous_new_list_token.indent_level

        return (
            starting_whitespace,
            did_container_start_midline,
            block_quote_leading_space_length,
            had_weird_block_quote_in_list,
            list_leading_space_length,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_post_adjust_whitespace(
        starting_whitespace: str,
        containing_list_token: ListStartMarkdownToken,
        block_quote_leading_space_length: int,
        list_leading_space_length: int,
        list_start_content_length: int,
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
    ) -> str:
        POGGER.debug(f"adj->starting_whitespace>>:{starting_whitespace}:<<")
        POGGER.debug(
            f"adj->containing_list_token.indent_level>>:{containing_list_token.indent_level}:<<"
        )
        POGGER.debug(
            f"adj->block_quote_leading_space_length>>:{block_quote_leading_space_length}:<<"
        )
        POGGER.debug(f"adj->list_leading_space_length>>:{list_leading_space_length}:<<")
        POGGER.debug(f"list_start_content_length:{list_start_content_length}:<<")

        pad_to_length = (
            current_token.column_number - 1
            if current_token.is_new_list_item
            else (
                containing_list_token.indent_level
                - block_quote_leading_space_length
                - list_leading_space_length
                - list_start_content_length
            )
        )
        POGGER.debug(f"pad_to_length:{pad_to_length}:<<")
        POGGER.debug(f"adj->starting_whitespace>>:{starting_whitespace}:<<")
        post_adjust_whitespace = starting_whitespace.ljust(pad_to_length, " ")
        POGGER.debug(f"adj->post_adjust_whitespace>>:{post_adjust_whitespace}:<<")
        return post_adjust_whitespace

    # pylint: enable=too-many-arguments

    @staticmethod
    def __rehydrate_list_start_contained_in_list_spacingx(
        containing_list_token: ListStartMarkdownToken,
        current_token: Union[NewListItemMarkdownToken, ListStartMarkdownToken],
        block_quote_leading_space_length: int,
    ) -> Tuple[int, str]:
        previous_indent = containing_list_token.indent_level
        white_space_length = (
            len(current_token.extracted_whitespace) + block_quote_leading_space_length
        )
        POGGER.debug(f"adj->len(ws)>>:{white_space_length}:<<")
        extracted_whitespace = (
            "".ljust(white_space_length - previous_indent, " ")
            if white_space_length > previous_indent
            else ""
        )
        POGGER.debug(f"adj->previous_indent>>:{previous_indent}:<<")
        POGGER.debug(
            f"adj->extracted_whitespace>>:{ParserHelper.make_value_visible(extracted_whitespace)}:<<"
        )
        return previous_indent, extracted_whitespace

    @staticmethod
    def __rehydrate_list_start_contained_in_list_deeper_block_quote(
        context: MarkdownTransformContext,
        previous_token: MarkdownToken,
        deeper_containing_block_quote_token: BlockQuoteMarkdownToken,
        current_token: Union[ListStartMarkdownToken, NewListItemMarkdownToken],
    ) -> Tuple[bool, str, bool, int, bool]:
        POGGER.debug(
            f"previous_token:{ParserHelper.make_value_visible(previous_token)}:"
        )
        # if previous_token.is_end_token:
        #     POGGER.debug(
        #         f"previous_token.start_markdown_token:{previous_token.start_markdown_token}:"
        #     )
        POGGER.debug(
            f"deeper_containing_block_quote_token:{ParserHelper.make_value_visible(deeper_containing_block_quote_token)}:"
        )
        had_weird_block_quote_in_list = False
        do_perform_block_quote_ending = False
        if previous_token.is_end_token:
            previous_end_token = cast(EndMarkdownToken, previous_token)
            if previous_end_token.start_markdown_token.is_block_quote_start:
                had_weird_block_quote_in_list = True
                POGGER.debug(f"previous_token:{previous_token}:")
                POGGER.debug(
                    f"previous_token.start_markdown_token:{previous_end_token.start_markdown_token}:"
                )
                block_quote_token = cast(
                    BlockQuoteMarkdownToken, previous_end_token.start_markdown_token
                )
                POGGER.debug(
                    f"previous_token.start_markdown_token.leading_spaces:{block_quote_token.bleading_spaces}:"
                )
                assert (
                    block_quote_token.bleading_spaces is not None
                ), "Bleading spaces must be defined by now."
                newline_count = ParserHelper.count_characters_in_text(
                    block_quote_token.bleading_spaces, "\n"
                )
                previous_start_line = block_quote_token.line_number
                POGGER.debug(f"newline_count:{newline_count}:")
                POGGER.debug(f"previous_start_line:{previous_start_line}:")
                projected_start_line = previous_start_line + (
                    newline_count + 1
                )  # 044lld off by 2  044lle off by 1
                if block_quote_token.weird_kludge_two:
                    projected_start_line += block_quote_token.weird_kludge_two
                POGGER.debug(f"projected_start_line:{projected_start_line}:")
                POGGER.debug(f"current_token.line_number:{current_token.line_number}:")
                do_perform_block_quote_ending = (
                    projected_start_line != current_token.line_number
                )
                # assert projected_start_line in [
                #     current_token.line_number,
                #     current_token.line_number + 1,
                # ], "should be one of the two, unless we have miscalculated"
        (
            block_quote_leading_space,
            starting_whitespace,
            did_container_start_midline,
            check_list_for_indent,
        ) = TransformListBlock.__rehydrate_list_start_deep(
            context,
            do_perform_block_quote_ending,
            previous_token,
            current_token,
            deeper_containing_block_quote_token,
            had_weird_block_quote_in_list,
        )
        POGGER.debug(
            f"block_quote_leading_space:{ParserHelper.make_value_visible(block_quote_leading_space)}:"
        )

        POGGER.debug(f"starting_whitespace:{starting_whitespace}:")
        return (
            check_list_for_indent,
            starting_whitespace,
            did_container_start_midline,
            len(block_quote_leading_space),
            had_weird_block_quote_in_list,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __rehydrate_list_start_deep(
        context: MarkdownTransformContext,
        do_perform_block_quote_ending: bool,
        previous_token: MarkdownToken,
        current_token: MarkdownToken,
        deeper_containing_block_quote_token: Optional[BlockQuoteMarkdownToken],
        had_weird_block_quote_in_list: bool,
    ) -> Tuple[str, str, bool, bool]:
        starting_whitespace = ""
        did_container_start_midline = False
        check_list_for_indent = True
        if do_perform_block_quote_ending:
            assert (
                previous_token.is_end_token
            ), "Block quote ending must indicate an end token."
            previous_end_token = cast(EndMarkdownToken, previous_token)
            POGGER.debug(
                f">>{ParserHelper.make_value_visible(previous_end_token.start_markdown_token)}"
            )
            previous_block_quote_token = cast(
                BlockQuoteMarkdownToken, previous_end_token.start_markdown_token
            )
            assert (
                previous_block_quote_token.bleading_spaces is not None
            ), "Bleading spaces must be defined by now."
            split_leading_spaces = previous_block_quote_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            POGGER.debug(f"split_leading_spaces>>{split_leading_spaces}")
            POGGER.debug(
                f"current_token>>{ParserHelper.make_value_visible(current_token)}"
            )
            # if (
            #     current_token.is_new_list_item
            #     and len(split_leading_spaces) <= 2
            #     and False
            # ):
            #     block_quote_leading_space = ""
            #     starting_whitespace = ""
            # else:

            block_quote_leading_space = split_leading_spaces[-1]
            starting_whitespace = block_quote_leading_space
            did_container_start_midline = True
            # up to here?
            check_list_for_indent = False
        else:
            assert (
                deeper_containing_block_quote_token is not None
            ), "This condition must exist for the logic to be here."
            POGGER.debug(
                f"adj->deeper_containing_block_quote_token.line_number>>:{deeper_containing_block_quote_token.line_number}:<<"
            )

            POGGER.debug(
                f"adj->current_token.line_number>>:{current_token.line_number}:<<"
            )
            # line_number_delta = ParserHelper.count_newlines_in_text(transformed_data) - current_token.line_number
            line_number_delta = (
                current_token.line_number
                - deeper_containing_block_quote_token.line_number
            )
            POGGER.debug(f"index:{line_number_delta}")
            assert deeper_containing_block_quote_token
            # if deeper_containing_block_quote_token:
            adjust_token_index = next(  # pragma: no cover
                (
                    i
                    for i in range(len(context.container_token_stack))
                    if context.container_token_stack[i]
                    == deeper_containing_block_quote_token
                ),
                None,
            )
            assert adjust_token_index is not None
            line_number_delta -= context.container_token_indents[
                adjust_token_index
            ].adjustment
            # endif

            assert (
                deeper_containing_block_quote_token.bleading_spaces is not None
            ), "Bleading spaces must be defined by now."
            split_leading_spaces = (
                deeper_containing_block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
            )
            POGGER.debug(
                f"split_leading_spaces:{ParserHelper.make_value_visible(split_leading_spaces)}"
            )

            block_quote_leading_space = split_leading_spaces[line_number_delta]
            if had_weird_block_quote_in_list:
                starting_whitespace = block_quote_leading_space
        return (
            block_quote_leading_space,
            starting_whitespace,
            did_container_start_midline,
            check_list_for_indent,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __rehydrate_list_start_calculate_start(
        current_token: ListStartMarkdownToken,
        next_token: MarkdownToken,
        extracted_whitespace: str,
        previous_indent: int,
        adjustment_since_newline: int,
        post_adjust_whitespace: Optional[str],
    ) -> str:
        start_sequence = (
            f"{extracted_whitespace}{current_token.list_start_sequence}"
            if current_token.is_unordered_list_start
            else f"{extracted_whitespace}{current_token.list_start_content}{current_token.list_start_sequence}"
        )
        POGGER.debug(f">>start_sequence>>:{start_sequence}:<<")
        old_start_sequence = start_sequence
        if next_token.is_blank_line:
            POGGER.debug("blank-line")
            POGGER.debug(f">>next_token.column_number>>:{next_token.column_number}:<<")
            POGGER.debug(
                f">>current_token.column_number>>:{current_token.column_number}:<<"
            )
            list_content_length = 1
            if not current_token.is_unordered_list_start:
                list_content_length += len(current_token.list_start_content)
            new_column_number = (
                next_token.column_number
                - current_token.column_number
                - list_content_length
            )
            start_sequence += ParserHelper.repeat_string(" ", new_column_number)
        elif next_token.is_list_end:
            POGGER.debug("list-end")
        else:
            POGGER.debug("not list-end and not blank-line")
            # POGGER.debug(
            #     f">>current_token>>:{ParserHelper.make_value_visible(current_token)}:<<"
            # )
            # POGGER.debug(
            #     f">>current_token.indent_level>>:{current_token.indent_level}:<<"
            # )
            POGGER.debug(f">>previous_indent>>:{previous_indent}:<<")
            POGGER.debug(f">>adjustment_since_newline>>:{adjustment_since_newline}:<<")
            requested_indent = (
                current_token.indent_level
                + len(extracted_whitespace)
                - (current_token.column_number - 1)
            )
            POGGER.debug(f">>requested_indent>>:{requested_indent}:<<")
            start_sequence = start_sequence.ljust(requested_indent, " ")
        POGGER.debug(
            f">>current_token>>:{ParserHelper.make_value_visible(current_token)}:<<"
        )
        POGGER.debug(f">>next_token>>:{ParserHelper.make_value_visible(next_token)}:<<")

        start_sequence = TransformListBlock.__rehydrate_list_start_calculate_start_calc(
            current_token,
            start_sequence,
            old_start_sequence,
            post_adjust_whitespace,
        )
        return start_sequence

    # pylint: enable=too-many-arguments

    @classmethod
    def __rehydrate_list_start_calculate_start_calc(
        cls,
        current_token: ListStartMarkdownToken,
        start_sequence: str,
        old_start_sequence: str,
        post_adjust_whitespace: Optional[str],
    ) -> str:
        POGGER.debug(f">>tabbed_adjust>>:{str(current_token.tabbed_adjust)}:<<")
        if current_token.tabbed_adjust >= 0:
            POGGER.debug(f">>start_sequence>>:{start_sequence}:<<")
            POGGER.debug(f">>old_start_sequence>>:{old_start_sequence}:<<")
            spaces_to_consume = current_token.tabbed_adjust + 1
            POGGER.debug(f">>spaces_to_consume>>:{str(spaces_to_consume)}:<<")
            start_sequence = (
                start_sequence[: len(old_start_sequence)]
                + "\t"
                + start_sequence[len(old_start_sequence) + spaces_to_consume :]
            )
            POGGER.debug(
                f">>start_sequence>>:{ParserHelper.make_value_visible(start_sequence)}:<<"
            )

        POGGER.debug(f"<<start_sequence<<:{start_sequence}:<<")
        if post_adjust_whitespace:
            POGGER.debug(
                f"<<post_adjust_whitespace<<(post):{post_adjust_whitespace}:<<"
            )
            start_sequence = post_adjust_whitespace + start_sequence
            POGGER.debug(f"<<start_sequence<<(post):{start_sequence}:<<")

        if current_token.tabbed_extracted_whitespace is not None:
            start_sequence = (
                current_token.tabbed_extracted_whitespace
                + start_sequence[len(current_token.extracted_whitespace) :]
            )
        return start_sequence



"""
Module to provide for an encapsulation of the block quote element.
"""

import logging
from typing import Dict, Optional, Tuple, Union

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.container_markdown_token import ContainerMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_gfm.transform_to_gfm_list_looseness import (
    TransformToGfmListLooseness,
)
from pymarkdown.transform_markdown.markdown_transform_context import (
    RegisterHtmlTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-instance-attributes


class BlockQuoteMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of the block quote element.
    """

    def __init__(
        self, extracted_whitespace: str, position_marker: PositionMarker
    ) -> None:
        self.__extracted_whitespace, self.__leading_spaces, self.leading_text_index = (
            extracted_whitespace,
            "",
            0,
        )
        self.__tabbed_leading_spaces: Dict[int, str] = {}
        ContainerMarkdownToken.__init__(
            self,
            MarkdownToken._token_block_quote,
            "",
            position_marker=position_marker,
        )
        self.__compose_extra_data_field()
        self.weird_kludge_one: Optional[int] = None
        self.weird_kludge_two: Optional[int] = None
        self.weird_kludge_three: bool = False
        self.weird_kludge_four: Optional[Tuple[int, int, int, str]] = None
        self.weird_kludge_five = False
        self.weird_kludge_six = False
        self.weird_kludge_seven = False

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_block_quote

    # pylint: enable=protected-access

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def bleading_spaces(self) -> Optional[str]:
        """
        Returns any leading spaces that preface the block quote.
        """
        return self.__leading_spaces

    @property
    def tabbed_bleading_spaces(self) -> Dict[int, str]:
        """
        Returns the tabbed information for any leading spaces
        """
        return self.__tabbed_leading_spaces

    def add_bleading_spaces(
        self,
        leading_spaces_to_add: str,
        skip_adding_newline: bool = False,
        tabbed_leading_spaces: Optional[str] = None,
    ) -> None:
        """
        Add any leading spaces to the token, separating them with line feeds.
        """
        POGGER.debug(
            "__leading_spaces>>:$:<<",
            self.__leading_spaces,
        )
        POGGER.debug("add_leading_spaces>>:$:<<", leading_spaces_to_add)
        self.__leading_spaces = (
            f"{self.__leading_spaces}{leading_spaces_to_add}"
            if skip_adding_newline
            else (
                f"{self.__leading_spaces}{ParserHelper.newline_character}{leading_spaces_to_add}"
                if self.__leading_spaces
                else leading_spaces_to_add
            )
        )
        self.weird_kludge_five = True
        POGGER.debug(
            "__leading_spaces>>:$:<<",
            self.__leading_spaces,
        )
        if not skip_adding_newline and tabbed_leading_spaces:
            POGGER.debug(
                "__tabbed_leading_spaces>>:$:<<",
                self.__tabbed_leading_spaces,
            )
            newline_count = ParserHelper.count_newlines_in_text(self.__leading_spaces)
            self.__tabbed_leading_spaces[newline_count] = tabbed_leading_spaces
            POGGER.debug(
                "__tabbed_leading_spaces>>:$:<<",
                self.__tabbed_leading_spaces,
            )
        self.__compose_extra_data_field()

    def remove_last_bleading_space(self) -> str:
        """
        Remove the last leading space and return it.
        """
        last_separator_index = self.__leading_spaces.rfind("\n")
        if last_separator_index == -1:
            extracted_text = self.__leading_spaces
            self.__leading_spaces = ""
            self.weird_kludge_five = False
        else:
            extracted_text = self.__leading_spaces[last_separator_index + 1 :]
            self.__leading_spaces = self.__leading_spaces[:last_separator_index]
        self.leading_text_index -= 1
        self.__compose_extra_data_field()
        return extracted_text

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        item_list = [self.__extracted_whitespace, self.__leading_spaces]
        if self.__tabbed_leading_spaces:
            item_list.append(
                ParserHelper.make_value_visible(self.__tabbed_leading_spaces).replace(
                    "'", '"'
                )
            )
        self._set_extra_data(MarkdownToken.extra_data_separator.join(item_list))

    def calculate_next_bleading_space_part(
        self, increment_index: bool = True, delta: int = 0, allow_overflow: bool = False
    ) -> str:
        """
        Calculate the next leading space based on the leading_text_index,
        optonally incrementing it as well.
        """
        assert (
            self.bleading_spaces is not None
        ), "Bleading spaces must be defined by now."

        # print(f"increment_index>>:{increment_index}:<<")
        tabbed_leading = None
        if increment_index and self.__tabbed_leading_spaces:
            tabbed_leading = self.__tabbed_leading_spaces[self.leading_text_index]
            # print(f"dg>>:{ParserHelper.make_value_visible(dg)}:<<")

        split_leading_spaces = self.bleading_spaces.split(
            ParserHelper.newline_character
        )
        absolute_index = self.leading_text_index + delta
        assert not (allow_overflow and absolute_index >= len(split_leading_spaces))
        # if allow_overflow and absolute_index >= len(split_leading_spaces):
        #     leading_text = ""
        # else:
        leading_text = split_leading_spaces[self.leading_text_index + delta]
        if increment_index:
            self.leading_text_index += 1

        if tabbed_leading is not None:
            leading_text = tabbed_leading

        return leading_text

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            BlockQuoteMarkdownToken,
            BlockQuoteMarkdownToken.__handle_start_block_quote_token,
            BlockQuoteMarkdownToken.__handle_end_block_quote_token,
        )

    @staticmethod
    def __handle_start_block_quote_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        token_parts = [output_html]
        if output_html and output_html[-1] != ParserHelper.newline_character:
            token_parts.append(ParserHelper.newline_character)
        transform_state.is_in_loose_list = True
        token_parts.extend(["<blockquote>", ParserHelper.newline_character])
        return "".join(token_parts)

    @staticmethod
    def __handle_end_block_quote_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        token_parts = [output_html]
        if output_html[-1] != ParserHelper.newline_character:
            token_parts.append(ParserHelper.newline_character)
        transform_state.is_in_loose_list = (
            TransformToGfmListLooseness.reset_list_looseness(
                transform_state.actual_tokens,
                transform_state.actual_token_index,
            )
        )
        token_parts.extend(["</blockquote>", ParserHelper.newline_character])
        return "".join(token_parts)

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "bleading_spaces" and isinstance(field_value, str):
            self.__leading_spaces = field_value
            self.__compose_extra_data_field()
            return True
        return super()._modify_token(field_name, field_value)


# pylint: enable=too-many-instance-attributes



"""
Module to provide for an element that can be added to markdown parsing stream.
"""

from enum import Enum
from typing import List, Optional, Union

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.plugin_manager.bad_plugin_fix_error import BadPluginFixError
from pymarkdown.plugin_manager.plugin_modify_context import PluginModifyContext


class MarkdownTokenClass(Enum):
    """
    Enumeration to provide guidance on what class of token the token is.
    """

    CONTAINER_BLOCK = 0
    LEAF_BLOCK = 1
    INLINE_BLOCK = 2
    SPECIAL = 3


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class MarkdownToken:
    """
    Class to provide for a base encapsulation of the markdown tokens.
    """

    extra_data_separator = ":"

    _end_token_prefix = "end-"

    _token_pragma = "pragma"
    _token_end_of_stream = "end-of-stream"

    _token_task_list = "task-list"
    _token_front_matter = "front-matter"

    _token_block_quote = "block-quote"
    _token_unordered_list_start = "ulist"
    _token_ordered_list_start = "olist"
    _token_new_list_item = "li"

    _token_paragraph = "para"
    _token_blank_line = "BLANK"
    _token_atx_heading = "atx"
    _token_setext_heading = "setext"
    _token_thematic_break = "tbreak"
    _token_link_reference_definition = "link-ref-def"
    _token_html_block = "html-block"
    _token_fenced_code_block = "fcode-block"
    _token_indented_code_block = "icode-block"
    _token_text = "text"
    _token_table_main = "table"
    _token_table_header = "table-header"
    _token_table_header_item = "table-header-item"
    _token_table_body = "table-body"
    _token_table_row = "table-row"
    _token_table_row_item = "table-row-item"

    _token_inline_code_span = "icode-span"
    _token_inline_hard_break = "hard-break"
    _token_inline_uri_autolink = "uri-autolink"
    _token_inline_email_autolink = "email-autolink"
    _token_inline_raw_html = "raw-html"
    _token_inline_emphasis = "emphasis"
    _token_inline_link = "link"
    _token_inline_image = "image"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_name: str,
        token_class: MarkdownTokenClass,
        extra_data: Optional[str] = None,
        line_number: int = 0,
        column_number: int = 0,
        position_marker: Optional[PositionMarker] = None,
        is_extension: bool = False,
        can_force_close: bool = True,
        requires_end_token: bool = False,
        is_special: bool = False,
    ):
        if position_marker:
            line_number, column_number = (
                position_marker.line_number,
                position_marker.index_number + position_marker.index_indent + 1,
            )
        (
            self.__token_name,
            self.__token_class,
            self.__extra_data,
            self.__line_number,
            self.__column_number,
            self.__is_extension,
            self.__requires_end_token,
            self.__can_force_close,
            self.__is_special,
        ) = (
            token_name,
            token_class,
            extra_data,
            line_number,
            column_number,
            is_extension,
            requires_end_token,
            can_force_close,
            is_special,
        )

    # pylint: enable=too-many-arguments

    def __str__(self) -> str:
        return self.debug_string()

    def debug_string(self, include_column_row_info: bool = True) -> str:
        """
        More customizable version of __str__ that allows for options.
        """
        debug_parts = ["[", self.token_name]
        if include_column_row_info and (self.line_number or self.column_number):
            debug_parts.append(f"({self.line_number},{self.column_number})")
        if self.extra_data or self.is_paragraph or self.is_blank_line:
            debug_parts.append(f":{self.extra_data}")
        debug_parts.append("]")
        return "".join(debug_parts)

    def __repr__(self) -> str:
        return f"'{self.__str__()}'"

    @property
    def token_name(self) -> str:
        """
        Returns the name associated with the token.
        """
        return self.__token_name

    @property
    def is_container(self) -> bool:
        """
        Returns whether the current token is a container block element.
        """
        return self.__token_class == MarkdownTokenClass.CONTAINER_BLOCK

    @property
    def is_leaf(self) -> bool:
        """
        Returns whether the current token is a leaf block element.
        """
        return self.__token_class == MarkdownTokenClass.LEAF_BLOCK

    @property
    def extra_data(self) -> Optional[str]:
        """
        Returns the extra data associated with the token.
        """
        return self.__extra_data

    def _set_extra_data(self, extra_data: Optional[str]) -> None:
        self.__extra_data = extra_data

    @property
    def line_number(self) -> int:
        """
        Returns the line number associated with the token.
        """
        return self.__line_number

    @property
    def column_number(self) -> int:
        """
        Returns the column number associated with the token.
        """
        return self.__column_number

    def _set_column_number(self, column_number: int) -> None:
        self.__column_number = column_number

    @property
    def is_extension(self) -> bool:
        """
        Returns whether this token is implemented as an extension.
        """
        return self.__is_extension

    @property
    def requires_end_token(self) -> bool:
        """
        Returns whether this token requires an end token to complete it.
        """
        return self.__requires_end_token

    @property
    def can_force_close(self) -> bool:
        """
        Returns whether this token can be forceably closed.
        """
        return self.__can_force_close

    @property
    def is_special(self) -> bool:
        """
        Returns whether the current token is actually a special subclass.
        """
        return bool(self.__is_special)

    @property
    def is_end_token(self) -> bool:
        """
        Returns whether the current token is an end element.
        """
        return bool(self.token_name.startswith(MarkdownToken._end_token_prefix))

    @property
    def is_block(self) -> bool:
        """
        Returns whether the current token is one of the block tokens.
        """
        return bool(
            self.is_block_quote_start
            or self.is_list_start
            or self.is_thematic_break
            or self.is_atx_heading
            or self.is_setext_heading
            or self.is_code_block
            or self.is_html_block
            or self.is_paragraph
        )
        # or tables

    @property
    def is_end_of_stream(self) -> bool:
        """
        Returns whether the current token is the end of stream element.
        """
        return self.token_name == MarkdownToken._token_end_of_stream

    @property
    def is_pragma(self) -> bool:
        """
        Returns whether the current token is the pragma element.
        """
        return self.token_name == MarkdownToken._token_pragma

    @property
    def is_blank_line(self) -> bool:
        """
        Returns whether the current token is the blank line element.
        """
        return self.token_name == MarkdownToken._token_blank_line

    @property
    def is_block_quote_start(self) -> bool:
        """
        Returns whether the current token is a block quote.
        """
        return self.token_name == MarkdownToken._token_block_quote

    @property
    def is_block_quote_end(self) -> bool:
        """
        Returns whether the current token is a block quote.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_block_quote
        )

    @property
    def is_list_start(self) -> bool:
        """
        Returns whether the current token is a list element.
        """
        return bool(self.is_unordered_list_start or self.is_ordered_list_start)

    @property
    def is_list_end(self) -> bool:
        """
        Returns whether the current token is a list end element.
        """
        return bool(self.is_unordered_list_end or self.is_ordered_list_end)

    @property
    def is_unordered_list_start(self) -> bool:
        """
        Returns whether the current token is an unordered list element.
        """
        return self.token_name == MarkdownToken._token_unordered_list_start

    @property
    def is_ordered_list_start(self) -> bool:
        """
        Returns whether the current token is an ordered list element.
        """
        return self.token_name == MarkdownToken._token_ordered_list_start

    @property
    def is_unordered_list_end(self) -> bool:
        """
        Returns whether the current token is an unordered list end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix
            + MarkdownToken._token_unordered_list_start
        )

    @property
    def is_ordered_list_end(self) -> bool:
        """
        Returns whether the current token is an ordered list end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_ordered_list_start
        )

    @property
    def is_container_end_token(self) -> bool:
        """
        Returns whether the current token is an end element for a container elements.
        """
        return bool(self.is_list_end or self.is_block_quote_end)

    @property
    def is_leaf_end_token(self) -> bool:
        """
        Returns whether the current token is an end element for a leaf element.
        """
        return bool(
            (
                self.is_paragraph_end
                or self.is_atx_heading_end
                or self.is_setext_heading_end
                or self.is_thematic_break
                or self.is_html_block_end
                or self.is_fenced_code_block_end
                or self.is_indented_code_block_end
            )
        )

    @property
    def is_new_list_item(self) -> bool:
        """
        Returns whether the current token is a list item element.
        """
        return self.token_name == MarkdownToken._token_new_list_item

    @property
    def is_any_list_token(self) -> bool:
        """
        Returns whether the current token is a list item element or a list element.
        """
        return bool(self.is_new_list_item or self.is_list_start)

    @property
    def is_paragraph(self) -> bool:
        """
        Returns whether the current token is a paragraph element.
        """
        return self.token_name == MarkdownToken._token_paragraph

    @property
    def is_paragraph_end(self) -> bool:
        """
        Returns whether the current token is a paragraph end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_paragraph
        )

    @property
    def is_thematic_break(self) -> bool:
        """
        Returns whether the current token is a thematic break element.
        """
        return self.token_name == MarkdownToken._token_thematic_break

    @property
    def is_front_matter(self) -> bool:
        """
        Returns whether the current token is the front matter element.
        """
        return self.token_name == MarkdownToken._token_front_matter

    @property
    def is_task_list(self) -> bool:
        """
        Returns whether the current token is the task list element.
        """
        return self.token_name == MarkdownToken._token_task_list

    @property
    def is_text(self) -> bool:
        """
        Returns whether the current token is a text element.
        """
        return self.token_name == MarkdownToken._token_text

    @property
    def is_special_text(self) -> bool:
        """
        Returns whether the current token is a special text element.
        """
        return bool(self.is_text and self.is_special)

    @property
    def is_setext_heading(self) -> bool:
        """
        Returns whether the current token is a setext heading element.
        """
        return self.token_name == MarkdownToken._token_setext_heading

    @property
    def is_setext_heading_end(self) -> bool:
        """
        Returns whether the current token is a setext heading end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_setext_heading
        )

    @property
    def is_atx_heading(self) -> bool:
        """
        Returns whether the current token is an atx element.
        """
        return self.token_name == MarkdownToken._token_atx_heading

    @property
    def is_atx_heading_end(self) -> bool:
        """
        Returns whether the current token is an atx heading end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_atx_heading
        )

    @property
    def is_table(self) -> bool:
        """
        Returns whether the current token is a table main element.
        """
        return self.token_name == MarkdownToken._token_table_main

    @property
    def is_table_end(self) -> bool:
        """
        Returns whether the current token is the end of a table element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_table_main
        )

    @property
    def is_table_header(self) -> bool:
        """
        Returns whether the current token is a header item from a table.
        """
        return self.token_name == self.token_name == MarkdownToken._token_table_header

    @property
    def is_table_header_item(self) -> bool:
        """
        Returns whether the current token is a header item from a table.
        """
        return (
            self.token_name == self.token_name == MarkdownToken._token_table_header_item
        )

    @property
    def is_table_row(self) -> bool:
        """
        Returns whether the current token is a row item from a table.
        """
        return self.token_name == self.token_name == MarkdownToken._token_table_row

    @property
    def is_table_row_item(self) -> bool:
        """
        Returns whether the current token is a row item from a table.
        """
        return self.token_name == self.token_name == MarkdownToken._token_table_row_item

    @property
    def is_code_block(self) -> bool:
        """
        Returns whether the current token is a code block element.
        """
        return bool(self.is_indented_code_block or self.is_fenced_code_block)

    @property
    def is_code_block_end(self) -> bool:
        """
        Returns whether the current token is a code block end element.
        """
        return bool(self.is_indented_code_block_end or self.is_fenced_code_block_end)

    @property
    def is_indented_code_block(self) -> bool:
        """
        Returns whether the current token is an indented code block element.
        """
        return self.token_name == MarkdownToken._token_indented_code_block

    @property
    def is_indented_code_block_end(self) -> bool:
        """
        Returns whether the current token is an indented code block end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix
            + MarkdownToken._token_indented_code_block
        )

    @property
    def is_fenced_code_block(self) -> bool:
        """
        Returns whether the current token is a fenced code block element.
        """
        return self.token_name == MarkdownToken._token_fenced_code_block

    @property
    def is_fenced_code_block_end(self) -> bool:
        """
        Returns whether the current token is a fenced code block element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_fenced_code_block
        )

    @property
    def is_link_reference_definition(self) -> bool:
        """
        Returns whether the current token is a link reference definition element.
        """
        return self.token_name == MarkdownToken._token_link_reference_definition

    @property
    def is_html_block(self) -> bool:
        """
        Returns whether the current token is a html block element.
        """
        return self.token_name == MarkdownToken._token_html_block

    @property
    def is_html_block_end(self) -> bool:
        """
        Returns whether the current token is a html block element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_html_block
        )

    @property
    def is_inline_code_span(self) -> bool:
        """
        Returns whether the current token is a code span element.
        """
        return self.token_name == MarkdownToken._token_inline_code_span

    @property
    def is_inline_hard_break(self) -> bool:
        """
        Returns whether the current token is a hard break element.
        """
        return self.token_name == MarkdownToken._token_inline_hard_break

    @property
    def is_inline_autolink(self) -> bool:
        """
        Returns whether the current token is an uri autolink or an email autolink element.
        """
        return bool(self.is_inline_uri_autolink or self.is_inline_email_autolink)

    @property
    def is_inline_uri_autolink(self) -> bool:
        """
        Returns whether the current token is an uri autolink element.
        """
        return self.token_name == MarkdownToken._token_inline_uri_autolink

    @property
    def is_inline_email_autolink(self) -> bool:
        """
        Returns whether the current token is an email autolink element.
        """
        return self.token_name == MarkdownToken._token_inline_email_autolink

    @property
    def is_inline_raw_html(self) -> bool:
        """
        Returns whether the current token is an email autolink element.
        """
        return self.token_name == MarkdownToken._token_inline_raw_html

    @property
    def is_inline_emphasis(self) -> bool:
        """
        Returns whether the current token is an emphasis element.
        """
        return self.token_name == MarkdownToken._token_inline_emphasis

    @property
    def is_inline_emphasis_end(self) -> bool:
        """
        Returns whether the current token is an emphasis end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_inline_emphasis
        )

    @property
    def is_inline_link(self) -> bool:
        """
        Returns whether the current token is a link element.
        """
        return self.token_name == MarkdownToken._token_inline_link

    @property
    def is_inline_link_end(self) -> bool:
        """
        Returns whether the current token is a link end element.
        """
        return (
            self.token_name
            == MarkdownToken._end_token_prefix + MarkdownToken._token_inline_link
        )

    @property
    def is_inline_image(self) -> bool:
        """
        Returns whether the current token is an image element.
        """
        return self.token_name == MarkdownToken._token_inline_image

    def adjust_line_number(
        self, context: PluginModifyContext, adjust_delta: int
    ) -> None:
        """
        Adjust the line number by a given amount.
        """
        # By design, tokens can only be modified in fix mode during the token pass.
        if not context.in_fix_mode:
            raise BadPluginFixError(
                f"Token '{self.__token_name}' can only be modified in fix mode."
            )
        if context.is_during_line_pass:
            raise BadPluginFixError(
                f"Token '{self.__token_name}' can only be modified during the token pass in fix mode."
            )
        if self.__line_number:
            self.__line_number += adjust_delta

    def modify_token(
        self,
        context: PluginModifyContext,
        field_name: str,
        field_value: Union[str, int],
    ) -> bool:
        """
        Within a given context, modify an existing token.
        """
        # By design, tokens can only be modified in fix mode during the token pass.
        if not context.in_fix_mode:
            raise BadPluginFixError(
                f"Token '{self.__token_name}' can only be modified in fix mode."
            )
        if context.is_during_line_pass:
            raise BadPluginFixError(
                f"Token '{self.__token_name}' can only be modified during the token pass in fix mode."
            )

        return self._modify_token(field_name, field_value)

    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "column_number" and isinstance(field_value, int):
            self.__column_number = field_value
            return True
        return False

    def generate_close_markdown_token_from_markdown_token(
        self,
        extracted_whitespace: str,
        extra_end_data: str,
        line_number: int = 0,
        column_number: int = 0,
    ) -> "MarkdownToken":
        """
        Generate the token emitted to close off the current stack token
        """
        return EndMarkdownToken(
            self.token_name,
            extracted_whitespace,
            extra_end_data,
            self,
            False,
            line_number=line_number,
            column_number=column_number,
        )

    @staticmethod
    def assert_tokens_are_same_except_for_line_number(
        token1: "MarkdownToken", token2: "MarkdownToken"
    ) -> None:
        """This assert function is needed as fixes to existing markdown tokens that
        have a open/close pairing may result in the open token being replaced with
        a new token that only differs by line number.
        """

        if str(token1) != str(token2):

            token1_visible = ParserHelper.make_value_visible(token1)
            token1_visible_first_index = token1_visible.index("(")
            token1_visible_second_index = token1_visible.index(
                ",", token1_visible_first_index
            )
            last_part = token1_visible[token1_visible_second_index:]
            first_part = token1_visible[: token1_visible_first_index + 1]
            fixed_token_text = f"{first_part}{token2.line_number}{last_part}"
            token2_visible = ParserHelper.make_value_visible(token2)

            assert (
                fixed_token_text == token2_visible
            ), f"{ParserHelper.make_value_visible(token1)}=={ParserHelper.make_value_visible(token2)}"


# pylint: enable=too-many-public-methods,too-many-instance-attributes


class EndMarkdownToken(MarkdownToken):
    """
    Class to provide for an encapsulation of the end element to a matching start.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        type_name: str,
        extracted_whitespace: str,
        extra_end_data: Optional[str],
        start_markdown_token: MarkdownToken,
        was_forced: bool,
        line_number: int = 0,
        column_number: int = 0,
    ) -> None:
        assert isinstance(start_markdown_token, MarkdownToken)
        assert (
            start_markdown_token.requires_end_token
        ), f"Token '{start_markdown_token} does not require end token."
        if not start_markdown_token.can_force_close:
            assert (
                not was_forced
            ), f"Token '{start_markdown_token}'s end token cannot be forced."

        (
            self.__type_name,
            self.__extracted_whitespace,
            self.__extra_end_data,
            self.__start_markdown_token,
            self.__was_forced,
        ) = (
            type_name,
            extracted_whitespace,
            extra_end_data,
            start_markdown_token,
            was_forced,
        )

        MarkdownToken.__init__(
            self,
            f"{MarkdownToken._end_token_prefix}{type_name}",
            MarkdownTokenClass.INLINE_BLOCK,
            "",
            line_number=line_number,
            column_number=column_number,
        )
        self.__compose_data_field()

    # pylint: enable=too-many-arguments

    @property
    def type_name(self) -> str:
        """
        Returns the type of markdown element related to this end element.
        """
        return self.__type_name

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def extra_end_data(self) -> Optional[str]:
        """
        Returns any extra data specifically tied to the end element.
        """
        return self.__extra_end_data

    def set_extra_end_data(self, new_end_data: Optional[str]) -> None:
        """
        Sets the extra data specifically tied to the end element. Use sparingly.
        """
        self.__extra_end_data = new_end_data
        self.__compose_data_field()

    @property
    def start_markdown_token(self) -> MarkdownToken:
        """
        Returns the start markdown token that this end token is the end for.
        """
        return self.__start_markdown_token

    @property
    def was_forced(self) -> bool:
        """
        Returns a value indicating whether the end element was forced.
        """
        return self.__was_forced

    def __compose_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        field_parts: List[str] = []
        if self.extra_end_data is not None:
            field_parts.extend((self.extracted_whitespace, self.extra_end_data))
        else:
            field_parts.extend(("", ""))
        if (
            isinstance(self.__start_markdown_token, MarkdownToken)
            and self.__start_markdown_token.can_force_close
        ):
            field_parts.append(str(self.was_forced))

        self._set_extra_data(MarkdownToken.extra_data_separator.join(field_parts))

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "extracted_whitespace" and isinstance(field_value, str):
            self.__extracted_whitespace = field_value
            self.__compose_data_field()
            return True
        if field_name == "extra_end_data" and isinstance(field_value, str):
            self.__extra_end_data = field_value
            self.__compose_data_field()
            return True
        return False



"""
Module to help with the parsing of bkactick inline elements.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineBacktickHelper:
    """
    Class to help with the parsing of bkactick inline elements.
    """

    code_span_bounds = "`"

    @staticmethod
    def handle_inline_backtick(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Handle the inline case of backticks for code spans.
        """
        _ = parser_properties

        POGGER.debug("before_collect>$", inline_request.next_index)
        POGGER.debug("inline_request.source_text>:$:<", inline_request.source_text)
        POGGER.debug("inline_request.tabified_text>:$:<", inline_request.tabified_text)
        (
            new_index,
            extracted_start_backticks,
        ) = ParserHelper.collect_while_one_of_characters_verified(
            inline_request.source_text,
            inline_request.next_index,
            InlineBacktickHelper.code_span_bounds,
        )
        POGGER.debug("after_collect>$>$", new_index, extracted_start_backticks)

        extracted_start_backticks_size, end_backtick_start_index = (
            len(extracted_start_backticks),
            inline_request.source_text.find(extracted_start_backticks, new_index),
        )
        while end_backtick_start_index != -1:
            (
                end_backticks_index,
                end_backticks_attempt,
            ) = ParserHelper.collect_while_one_of_characters_verified(
                inline_request.source_text,
                end_backtick_start_index,
                InlineBacktickHelper.code_span_bounds,
            )
            if len(end_backticks_attempt) == extracted_start_backticks_size:
                break
            end_backtick_start_index = inline_request.source_text.find(
                extracted_start_backticks, end_backticks_index
            )

        inline_response = InlineBacktickHelper.__build_backtick_response(
            inline_request,
            end_backtick_start_index,
            extracted_start_backticks,
            new_index,
            extracted_start_backticks_size,
        )

        assert (
            inline_response.new_index is not None
        ), "new_index should be defined by this point."
        POGGER.debug(
            ">>delta_line_number>>$<<",
            inline_response.delta_line_number,
        )
        POGGER.debug(
            ">>delta_column_number>>$<<",
            inline_response.delta_column_number,
        )
        if inline_response.delta_line_number == -1:
            inline_response.delta_line_number, inline_response.delta_column_number = (
                0,
                inline_response.new_index - inline_request.next_index,
            )
        return inline_response

    @staticmethod
    def __build_backtick_response(
        inline_request: InlineRequest,
        end_backtick_start_index: int,
        extracted_start_backticks: str,
        new_index: int,
        extracted_start_backticks_size: int,
    ) -> InlineResponse:
        inline_response = InlineResponse()
        inline_response.delta_line_number = -1

        if end_backtick_start_index == -1:
            inline_response.new_string, inline_response.new_index = (
                extracted_start_backticks,
                new_index,
            )
        else:
            (
                between_text,
                original_between_text,
                leading_whitespace,
                trailing_whitespace,
                inline_response.adj_newlines,
            ) = InlineBacktickHelper.__calculate_backtick_between_text(
                inline_request, new_index, end_backtick_start_index
            )

            POGGER.debug("between_text>>$<<", between_text)
            between_text = InlineHelper.append_text("", between_text)
            POGGER.debug("between_text>>$<<", between_text)
            POGGER.debug(
                "leading_whitespace>>$<<",
                leading_whitespace,
            )
            POGGER.debug(
                "trailing_whitespace>>$<<",
                trailing_whitespace,
            )
            end_backtick_start_index += extracted_start_backticks_size
            inline_response.new_string, inline_response.new_index = (
                "",
                end_backtick_start_index,
            )
            assert (
                inline_request.line_number is not None
            ), "line_number must be defined by now."
            assert (
                inline_request.column_number is not None
            ), "column_number must be defined by now."
            assert (
                inline_request.remaining_line is not None
            ), "remaining_line must be defined by now."
            new_column_number = inline_request.column_number + len(
                inline_request.remaining_line
            )
            POGGER.debug(">>new_column_number>>$", new_column_number)

            inline_response.new_tokens = [
                InlineCodeSpanMarkdownToken(
                    between_text,
                    extracted_start_backticks,
                    leading_whitespace,
                    trailing_whitespace,
                    inline_request.is_in_table,
                    inline_request.line_number,
                    new_column_number,
                )
            ]

            if ParserHelper.newline_character in original_between_text:
                (
                    inline_response.delta_line_number,
                    inline_response.delta_column_number,
                ) = ParserHelper.calculate_deltas(
                    f"{original_between_text}{extracted_start_backticks}"
                )
        return inline_response

    @staticmethod
    def __backtick_split_lines(input_string: str) -> List[str]:
        current_index = 0
        next_index, extracted_text = (
            ParserHelper.collect_while_one_of_characters_verified(
                input_string, current_index, " \t"
            )
        )
        split_array: List[str] = [extracted_text]
        # POGGER.debug("next_index:$, extracted_text>:$:<", next_index, extracted_text)
        while next_index < len(input_string):
            (
                current_index,
                extracted_text,
            ) = ParserHelper.collect_until_one_of_characters_verified(
                input_string, next_index, " \t"
            )
            split_array.append(extracted_text)
            next_index, extracted_text = (
                ParserHelper.collect_while_one_of_characters_verified(
                    input_string, current_index, " \t"
                )
            )
            split_array.append(extracted_text)
        return split_array

    @staticmethod
    def __adjust_for_injected_noops(between_text: str) -> str:

        prefix = (
            ParserHelper.escape_character
            + "\a"
            + ParserHelper.escape_character
            + ParserHelper.replace_noop_character
            + ParserHelper.escape_character
            + "\a"
        )
        suffix = ParserHelper.escape_character + "\a"
        start_index = 0
        search_index = between_text.find(prefix, start_index)
        if search_index != -1:
            string_parts = []
            while search_index != -1:
                string_parts.append(between_text[start_index:search_index])
                search_index += len(prefix)
                end_index = between_text.find(suffix, search_index)
                in_between_text = between_text[search_index:end_index]
                string_parts.append(
                    ParserHelper.create_replacement_markers(
                        ParserHelper.replace_noop_character, in_between_text
                    )
                )
                start_index = end_index + len(suffix)
                search_index = between_text.find(prefix, start_index)
            if start_index != len(between_text):
                string_parts.append(between_text[start_index:])
            between_text = "".join(string_parts)
        return between_text

    @staticmethod
    def __calculate_backtick_between_text(
        inline_request: InlineRequest, new_index: int, end_backtick_start_index: int
    ) -> Tuple[str, str, str, str, int]:
        POGGER.debug("inline_request.source_text>>$<<", inline_request.source_text)
        POGGER.debug("new_index>>$<<", new_index)
        POGGER.debug("end_backtick_start_index>>$<<", end_backtick_start_index)

        between_text = inline_request.source_text[new_index:end_backtick_start_index]

        original_between_text, leading_whitespace, trailing_whitespace = (
            between_text,
            "",
            "",
        )

        if inline_request.tabified_text:
            between_text, adj_newlines = (
                InlineBacktickHelper.__calculate_backtick_between_tabified_text(
                    inline_request, new_index, end_backtick_start_index
                )
            )
        else:
            adj_newlines = 0
        POGGER.debug(
            "after_collect>$>>$>>$<<",
            between_text,
            end_backtick_start_index,
            inline_request.source_text[end_backtick_start_index:],
        )
        POGGER.debug("between_text>>$<<", between_text)
        if (
            len(between_text) > 2
            and between_text[0]
            in [
                ParserHelper.space_character,
                ParserHelper.newline_character,
            ]
            and between_text[-1]
            in [
                ParserHelper.space_character,
                ParserHelper.newline_character,
            ]
        ):
            stripped_between_attempt = between_text[1:-1]
            if len(stripped_between_attempt.strip(" ")) != 0:
                leading_whitespace, trailing_whitespace = (
                    between_text[0],
                    between_text[-1],
                )
                between_text = stripped_between_attempt

        replaced_newline = ParserHelper.create_replacement_markers(
            ParserHelper.newline_character, ParserHelper.space_character
        )
        POGGER.debug("between_text>>$<<", between_text)
        between_text = ParserHelper.escape_special_characters(between_text)
        POGGER.debug("between_text>>$<<", between_text)
        POGGER.debug(
            "leading_whitespace>>$<<",
            leading_whitespace,
        )
        POGGER.debug(
            "trailing_whitespace>>$<<",
            trailing_whitespace,
        )

        between_text = InlineBacktickHelper.__adjust_for_injected_noops(between_text)

        between_text, leading_whitespace, trailing_whitespace = (
            between_text.replace(ParserHelper.newline_character, replaced_newline),
            leading_whitespace.replace(
                ParserHelper.newline_character, replaced_newline
            ),
            trailing_whitespace.replace(
                ParserHelper.newline_character, replaced_newline
            ),
        )
        return (
            between_text,
            original_between_text,
            leading_whitespace,
            trailing_whitespace,
            adj_newlines,
        )

    @staticmethod
    def __find_adjusted_index(
        source_text: str, new_index: int, adj_tabified_text: str, whitespace_to_add: str
    ) -> Tuple[int, int]:

        before_index_text = source_text[:new_index]
        # w2 = source_text[new_index:]
        before_index_newline_count = ParserHelper.count_newlines_in_text(
            before_index_text
        )
        if before_index_newline_count:
            last_newline_index = source_text.rindex("\n", 0, new_index) + 1
            adj_new_index = new_index - last_newline_index
        else:
            last_newline_index = 0
            adj_new_index = new_index

        # q4 = source_text.find("\n", last_newline_index)
        # if q4 == -1:
        #     q4 = len(source_text)
        # st_after_newline = source_text[last_newline_index:q4]

        tabified_start_index = (
            ParserHelper.find_nth_occurrence(
                adj_tabified_text, "\n", before_index_newline_count
            )
            + 1
        )
        # tabified_end_index = adj_tabified_text.find("\n", tabified_start_index)
        # if tabified_end_index == -1:
        #     tabified_end_index = len(adj_tabified_text)
        # tt_after_newline = adj_tabified_text[q2:q3]

        whitespace_start_index = (
            ParserHelper.find_nth_occurrence(
                whitespace_to_add, "\n", before_index_newline_count
            )
            + 1
        )
        whitespace_end_index = whitespace_to_add.find("\n", whitespace_start_index)
        if whitespace_end_index == -1:
            whitespace_end_index = len(adj_tabified_text)
        whitespace_after_newline = whitespace_to_add[
            whitespace_start_index:whitespace_end_index
        ]
        return (
            tabified_start_index + len(whitespace_after_newline) + adj_new_index,
            before_index_newline_count,
        )

    @staticmethod
    def __calc_tabified_text(
        inline_request: InlineRequest, new_index: int, end_backtick_start_index: int
    ) -> Tuple[Optional[str], Optional[str], bool, int]:
        adj_newlines = 0
        adj_tabified_text = inline_request.tabified_text
        normal_backtick_text = inline_request.source_text[
            new_index:end_backtick_start_index
        ]
        whitespace_to_add = inline_request.whitespace_to_recombine
        did_use_para_space = False
        if (
            "\n" in normal_backtick_text
            and inline_request.para_space
            and whitespace_to_add is None
        ):
            did_use_para_space = True
            whitespace_to_add = inline_request.para_space
        if "\n" in normal_backtick_text and whitespace_to_add is not None:
            assert adj_tabified_text is not None
            adj_tabified_text_newlines = ParserHelper.count_newlines_in_text(
                adj_tabified_text
            )
            whitespace_to_add_newlines = ParserHelper.count_newlines_in_text(
                whitespace_to_add
            )
            assert adj_tabified_text_newlines == whitespace_to_add_newlines

            split_adj_tabified_text = adj_tabified_text.split("\n")
            split_whitespace_to_add = whitespace_to_add.split("\n")
            for split_index, split_item in enumerate(split_adj_tabified_text):
                ffff = split_whitespace_to_add[split_index].startswith(
                    " "
                ) and split_whitespace_to_add[split_index].endswith("\t")
                split_whitespace_to_add[split_index] = (
                    split_item
                    if ffff
                    else split_whitespace_to_add[split_index] + split_item
                )
            adj_tabified_text = "\n".join(split_whitespace_to_add)
            adj_newlines = ParserHelper.count_newlines_in_text(normal_backtick_text)
        return adj_tabified_text, whitespace_to_add, did_use_para_space, adj_newlines

    @staticmethod
    def __calc_adjusted_between_text_kludge(
        inline_request: InlineRequest,
        actual_between_text: str,
        whitespace_to_add: str,
        start_line_index: int,
        end_line_index: int,
    ) -> str:
        assert inline_request.last_container_token is not None
        if inline_request.last_container_token.is_block_quote_start:
            last_block_token = cast(
                BlockQuoteMarkdownToken, inline_request.last_container_token
            )
            assert last_block_token.bleading_spaces is not None
            split_spaces = last_block_token.bleading_spaces.split("\n")
        else:
            last_list_token = cast(
                ListStartMarkdownToken, inline_request.last_container_token
            )
            assert last_list_token.leading_spaces is not None
            split_spaces = last_list_token.leading_spaces.split("\n")
        split_whitespace = whitespace_to_add.split("\n")
        # assert len(er1) == len(er2)
        for index, split_whitespace_item in enumerate(split_whitespace):
            split_spaces_item = split_spaces[index]
            if (
                split_whitespace_item == "\t"
                and not split_spaces_item
                and index > start_line_index
                and index <= end_line_index
            ):
                newline_index = (
                    ParserHelper.find_nth_occurrence(
                        actual_between_text, "\n", index - start_line_index
                    )
                    + 1
                )
                prefix = actual_between_text[:newline_index]
                suffix = actual_between_text[newline_index + 1 :]
                noop_part = ParserHelper.create_replacement_markers(
                    ParserHelper.replace_noop_character, "\t"
                )
                actual_between_text = prefix + noop_part + suffix
        return actual_between_text

    @staticmethod
    def __calc_adjusted_between_text(
        adj_newlines: int,
        inline_request: InlineRequest,
        start_index: int,
        adj_tabified_text: str,
        whitespace_to_add: str,
        end_backtick_start_index: int,
        did_use_para_space: bool,
    ) -> Tuple[str, int]:
        adj_start_index, start_line_index = InlineBacktickHelper.__find_adjusted_index(
            inline_request.source_text,
            start_index,
            adj_tabified_text,
            whitespace_to_add,
        )
        adj_end_index, end_line_index = InlineBacktickHelper.__find_adjusted_index(
            inline_request.source_text,
            end_backtick_start_index,
            adj_tabified_text,
            whitespace_to_add,
        )
        actual_between_text = adj_tabified_text[adj_start_index:adj_end_index]
        if did_use_para_space and inline_request.last_container_token:
            actual_between_text = (
                InlineBacktickHelper.__calc_adjusted_between_text_kludge(
                    inline_request,
                    actual_between_text,
                    whitespace_to_add,
                    start_line_index,
                    end_line_index,
                )
            )
        return actual_between_text, adj_newlines

    @staticmethod
    def __calculate_backtick_between_tabified_text(
        inline_request: InlineRequest, new_index: int, end_backtick_start_index: int
    ) -> Tuple[str, int]:
        POGGER.debug("inline_request.tabified_text>>$<<", inline_request.tabified_text)
        split_source_lines = InlineBacktickHelper.__backtick_split_lines(
            inline_request.source_text
        )
        POGGER.debug("split_source_lines>>$<<", split_source_lines)
        assert (
            inline_request.tabified_text is not None
        ), "tabified_text must be defined by now."
        adj_tabified_text, whitespace_to_add, did_use_para_space, adj_newlines = (
            InlineBacktickHelper.__calc_tabified_text(
                inline_request, new_index, end_backtick_start_index
            )
        )
        assert adj_tabified_text is not None
        split_tabified_lines = InlineBacktickHelper.__backtick_split_lines(
            adj_tabified_text
        )
        if len(split_tabified_lines) != len(split_source_lines):
            assert whitespace_to_add is not None
            return InlineBacktickHelper.__calc_adjusted_between_text(
                adj_newlines,
                inline_request,
                new_index,
                adj_tabified_text,
                whitespace_to_add,
                end_backtick_start_index,
                did_use_para_space,
            )

        POGGER.debug("split_tabified_lines>>$<<", split_tabified_lines)

        (
            start_array_index,
            start_delta,
            calculated_index,
        ) = InlineBacktickHelper.__find_index_in_split_lines(
            split_source_lines, new_index
        )
        POGGER.debug(
            "start_array_index=$, start_delta=$, calculated_index=$",
            start_array_index,
            start_delta,
            calculated_index,
        )
        assert (
            calculated_index + start_delta == new_index
        ), "calculations should equal current position"

        (
            end_array_index,
            end_delta,
            calculated_index,
        ) = InlineBacktickHelper.__find_index_in_split_lines(
            split_source_lines, end_backtick_start_index
        )
        POGGER.debug(
            "end_array_index=$, end_delta=$, calculated_index=$",
            end_array_index,
            end_delta,
            calculated_index,
        )
        assert (
            calculated_index + end_delta == end_backtick_start_index
        ), "calculations should equal current position"

        if start_array_index == end_array_index:
            POGGER.debug("same")
            actual_between_text = split_tabified_lines[start_array_index][
                start_delta:end_delta
            ]
        else:
            POGGER.debug("borders")
            if start_delta:
                actual_between_text = split_tabified_lines[start_array_index][
                    start_delta:
                ]
                start_array_index += 1
            else:
                actual_between_text = ""
            actual_between_text += "".join(
                split_tabified_lines[i]
                for i in range(start_array_index, end_array_index)
            )
            if end_delta:
                actual_between_text += split_tabified_lines[end_array_index][:end_delta]
        POGGER.debug("actual_between_text>:$:<", actual_between_text)
        return actual_between_text, adj_newlines

    @staticmethod
    def __find_index_in_split_lines(
        split_array: List[str], index_to_find: int
    ) -> Tuple[int, int, int]:
        start_index = 0
        _array_index = 0
        for _array_index, array_element in enumerate(split_array):  # pragma: no cover
            POGGER.debug(
                "$--$ >>:$:<<", _array_index, start_index, split_array[_array_index]
            )
            if start_index <= index_to_find < start_index + len(array_element):
                break
            start_index += len(array_element)
        assert start_index != (
            start_index + len(split_array)
        ), "For loop must stop before the end."
        delta = index_to_find - start_index
        # d_before = split_array[_array_index][:delta]
        # d_after = split_array[_array_index][delta:]
        POGGER.debug("i=$,start_index=$,delta=$", _array_index, start_index, delta)
        return _array_index, delta, start_index



"""
Module to provide for the handlers for tokens to allow transformation into HTML.
"""

import inspect
import logging
from typing import Dict, List, Optional, cast

from pymarkdown.extensions.extension_token_types import ExtensionTokenTypes
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.token_types import TokenTypes
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    EndHtmlTokenTransformProtocol,
    StartHtmlTokenTransformProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods


class TransformToGfmTokenHandlers:
    """
    Class to provide for the handlers for tokens to allow transformation into HTML.
    """

    def __init__(self) -> None:
        self.__start_token_handlers: Dict[str, StartHtmlTokenTransformProtocol] = {}
        self.__end_token_handlers: Dict[str, EndHtmlTokenTransformProtocol] = {}

        ert = TokenTypes.get_inline_token_types()
        ert.extend(TokenTypes.get_leaf_token_types())
        ert.extend(TokenTypes.get_container_token_types())
        ert.extend(TokenTypes.get_special_token_types())
        ert.extend(ExtensionTokenTypes.get_token_types())

        for token_type in ert:
            token_init_fn = token_type.__dict__["__init__"]
            init_parameters = {
                i: "" for i in inspect.getfullargspec(token_init_fn)[0] if i != "self"
            }
            handler_instance = token_type(**init_parameters)
            handler_instance.register_for_html_transform(self.__register_handlers)

    def __register_handlers(
        self,
        token_type: type,
        start_token_handler: StartHtmlTokenTransformProtocol,
        end_token_handler: Optional[EndHtmlTokenTransformProtocol] = None,
    ) -> None:
        """
        Register the handlers necessary to deal with token's start and end.
        """
        assert issubclass(
            token_type, MarkdownToken
        ), f"Token class '{token_type}' must be descended from the 'MarkdownToken' class."

        # This is being done to create a rough "default" instance of the type in
        # question.  Because it is a rough creation and we have no intentions of
        # using it, each parameter is an empty string.  Therefore, we expect that
        # mypy will have issues when we create the instance, hence ignoring them.
        #
        # That was the old way, and had problems.  The new way is to call a new
        # static method directly.  Both are here until that is completed.

        # if "get_markdown_token_type" in token_type.__dict__:
        token_name = token_type.__dict__["get_markdown_token_type"].__func__()
        # else:
        #     token_init_fn = token_type.__dict__["__init__"]
        #     init_parameters = {
        #         i: "" for i in inspect.getfullargspec(token_init_fn)[0] if i != "self"
        #     }
        #     handler_instance = token_type(**init_parameters)  # type: ignore
        #     token_name = handler_instance.token_name

        assert token_name, "This must not be empty."
        self.__start_token_handlers[token_name] = start_token_handler
        if end_token_handler:
            self.__end_token_handlers[token_name] = end_token_handler

    # pylint: disable=too-many-arguments
    def apply_transformation(
        self,
        transform_state: TransformState,
        actual_tokens: List[MarkdownToken],
        actual_tokens_size: int,
        next_token: MarkdownToken,
        output_html: str,
    ) -> str:
        """
        Apply the required tranformation for the current token.
        """
        transform_state.add_trailing_text = None
        transform_state.add_leading_text = None
        transform_state.next_token = None

        if (transform_state.actual_token_index + 1) < actual_tokens_size:
            transform_state.next_token = actual_tokens[
                transform_state.actual_token_index + 1
            ]
        if next_token.token_name in self.__start_token_handlers:
            start_handler_fn = self.__start_token_handlers[next_token.token_name]
            POGGER.debug("next_token>:$:<", next_token)
            POGGER.debug("output_html>:$:<", output_html)
            output_html = start_handler_fn(output_html, next_token, transform_state)
            POGGER.debug("output_html>:$:<", output_html)

        elif next_token.is_end_token:
            end_token = cast(EndMarkdownToken, next_token)
            if end_token.type_name not in self.__end_token_handlers:
                raise AssertionError(
                    f"Markdown token end type {end_token.type_name} not supported."
                )
            end_handler_fn = self.__end_token_handlers[end_token.type_name]
            POGGER.debug("end_token>:$:<", end_token)
            POGGER.debug("output_html>:$:<", output_html)
            output_html = end_handler_fn(output_html, end_token, transform_state)
            POGGER.debug("output_html>:$:<", output_html)
        else:
            raise AssertionError(
                f"Markdown token type {type(next_token)} not supported."
            )
        return output_html

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that ensures that top-level lists are surrounded by Blank Lines.

Note:  Because of the "weird" way in which this rule triggers, it must provide its own
       `override_is_error_token_prefaced_by_blank_line` logic, as modifying the normal
       logic for this one rule was not feasible.
"""

from typing import List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd032(RulePlugin):
    """
    Class to implement a plugin that ensures that top-level lists are surrounded by Blank Lines.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__last_non_end_token: Optional[MarkdownToken] = None
        self.__container_token_stack: List[MarkdownToken] = []
        self.__end_list_end_token: Optional[MarkdownToken] = None
        self.__previous_tokens: List[MarkdownToken] = []

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="blanks-around-lists",
            plugin_id="MD032",
            plugin_enabled_by_default=True,
            plugin_description="Lists should be surrounded by blank lines",
            plugin_version="0.5.1",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md032.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_non_end_token = None
        self.__container_token_stack.clear()
        self.__end_list_end_token = None
        self.__previous_tokens.clear()

    def __next_token_list_start(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        if (
            self.__last_non_end_token
            and not self.__last_non_end_token.is_blank_line
            and not (
                self.__container_token_stack
                and (
                    self.__container_token_stack[-1].is_list_start
                    or (
                        self.__container_token_stack[-1].is_block_quote_start
                        and self.__container_token_stack[-1].line_number
                        == token.line_number
                    )
                )
            )
        ):
            self.report_next_token_error(context, token)

        self.__container_token_stack.append(token)

    def __calculate_override_for_blank_line(
        self,
        token: MarkdownToken,
        token_index: int,
        parent_leaf_or_container_token: MarkdownToken,
    ) -> bool:
        # If we run into a blank line, there is a possibility that the blank line is followed
        # by a single line of text and then the HTML block end.  In that case, we need to
        # enable the override.
        #
        # i.e With the following markdown, we would get here via line 6 (second <script>),
        #     l0 would be line 4 (the blank line), and there is a text line (line 5) with
        #     an HTML block end and a list end that were skipped over.
        #
        # 1. a list
        #    <script>
        # <!-- pyml disable-next-line blanks-around-lists -->
        #
        #    </script>
        # <script>
        # </script>

        are_line_numbers_same = (
            parent_leaf_or_container_token.line_number == token.line_number - 2
        )
        is_eligible_text_token = (
            self.__previous_tokens[token_index + 1].line_number == token.line_number - 1
            and self.__previous_tokens[token_index + 1].is_text
        )
        is_html_end_present = self.__previous_tokens[token_index + 2].is_html_block_end
        return are_line_numbers_same and is_eligible_text_token and is_html_end_present

    def __calculate_override_for_fenced_code_block(self, token_index: int) -> bool:
        expected_fcb_text_token = self.__previous_tokens[token_index + 1]
        if expected_fcb_text_token.is_text:
            text_token = cast(TextMarkdownToken, expected_fcb_text_token)
            return not text_token.token_text or text_token.token_text.endswith(
                f"{ParserHelper.newline_character}{ParserHelper.replace_noop_character}"
            )
        return False

    def __calculate_override(self, token: MarkdownToken) -> bool:

        token_index = len(self.__previous_tokens) - 1
        while (
            token_index >= 0
            and not self.__previous_tokens[token_index].is_leaf
            and not self.__previous_tokens[token_index].is_container
        ):
            token_index -= 1

        parent_leaf_or_container_token = self.__previous_tokens[token_index]
        if parent_leaf_or_container_token.is_blank_line:
            return self.__calculate_override_for_blank_line(
                token, token_index, parent_leaf_or_container_token
            )
        if parent_leaf_or_container_token.is_fenced_code_block:
            return self.__calculate_override_for_fenced_code_block(token_index)

        assert self.__previous_tokens[token_index].is_leaf
        token_index -= 1
        if self.__previous_tokens[token_index].is_container:
            token_index -= 1
        base_token = self.__previous_tokens[token_index]

        is_base_token_blank = base_token.is_blank_line
        are_parent_and_token_on_same_line = (
            parent_leaf_or_container_token.line_number == token.line_number - 1
        )
        is_base_at_right_place = base_token.line_number == token.line_number - 2
        return (
            is_base_token_blank
            and are_parent_and_token_on_same_line
            and is_base_at_right_place
        )

    def __report_token(self, context: PluginScanContext, token: MarkdownToken) -> None:

        override_is_error_token_prefaced_by_blank_line = self.__calculate_override(
            token
        )

        line_number_delta = -1
        while context.is_pragma_on_line(token.line_number + line_number_delta):
            line_number_delta -= 1
        self.report_next_token_error(
            context,
            token,
            line_number_delta=line_number_delta,
            override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if self.__end_list_end_token:
            if (
                not token.is_blank_line
                and not token.is_new_list_item
                and not token.is_list_end
                and not token.is_block_quote_end
                and not token.is_end_of_stream
            ):
                self.__report_token(context, token)
            self.__end_list_end_token = None

        if token.is_block_quote_start:
            self.__container_token_stack.append(token)
        elif token.is_block_quote_end:
            del self.__container_token_stack[-1]
        elif token.is_list_start:
            self.__next_token_list_start(context, token)
        elif token.is_list_end:
            assert self.__last_non_end_token is not None
            if not self.__last_non_end_token.is_blank_line:
                self.__end_list_end_token = token
                del self.__container_token_stack[-1]

        if (
            not token.is_end_token
            and not token.is_block_quote_start
            and not token.is_list_start
        ):
            self.__last_non_end_token = token

        self.__previous_tokens.append(token)



"""
Module to help with the parsing of inline elements.
"""

import logging
from typing import Dict, List, Optional, Tuple

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_backslash_helper import InlineBackslashHelper
from pymarkdown.inline.inline_request import InlineRequest

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineHelper:
    """
    Class to help with the parsing of inline elements.
    """

    __html_character_escape_map = {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
        '"': "&quot;",
    }

    @staticmethod
    def extract_bounded_string(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        new_index: int,
        close_character: str,
        start_character: Optional[str],
    ) -> Tuple[int, Optional[str]]:
        """
        Extract a string that is bounded by some manner of characters.
        """
        nesting_level: int = 0
        break_characters = (
            f"{InlineBackslashHelper.backslash_character}{close_character}{start_character}"
            if start_character
            else f"{InlineBackslashHelper.backslash_character}{close_character}"
        )

        POGGER.debug(
            "extract_bounded_string>>new_index>>$>>data>>$>>",
            new_index,
            source_text[new_index:],
        )
        next_index, data = ParserHelper.collect_until_one_of_characters_verified(
            source_text, new_index, break_characters
        )
        extracted_parts: List[str] = [data]
        POGGER.debug(
            ">>next_index1>>$>>data>>$>>",
            next_index,
            data,
        )
        while next_index < len(source_text) and (
            source_text[next_index] != close_character or nesting_level != 0
        ):
            (
                next_index,
                nesting_level,
            ) = InlineHelper.__handle_next_extract_bounded_string_item(
                parser_properties,
                source_text,
                next_index,
                extracted_parts,
                start_character,
                nesting_level,
                close_character,
                break_characters,
            )
            POGGER.debug(
                "back>>next_index>>$>>data>>$>>",
                next_index,
                data,
            )
        POGGER.debug(
            ">>next_index2>>$>>data>>$>>",
            next_index,
            data,
        )
        if (
            ParserHelper.is_character_at_index(source_text, next_index, close_character)
            and nesting_level == 0
        ):
            POGGER.debug("extract_bounded_string>>found-close")
            return next_index + 1, "".join(extracted_parts)
        POGGER.debug(
            "extract_bounded_string>>ran out of string>>next_index>>$", next_index
        )
        return next_index, None

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_next_extract_bounded_string_item(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        next_index: int,
        extracted_parts: List[str],
        start_character: Optional[str],
        nesting_level: int,
        close_character: str,
        break_characters: str,
    ) -> Tuple[int, int]:
        if ParserHelper.is_character_at_index(
            source_text, next_index, InlineBackslashHelper.backslash_character
        ):
            POGGER.debug("pre-back>>next_index>>$>>", next_index)
            old_index = next_index

            inline_request = InlineRequest(source_text, next_index)
            inline_response = InlineBackslashHelper.handle_inline_backslash(
                parser_properties, inline_request
            )
            assert inline_response.new_index is not None, "New index must be defined."
            next_index = inline_response.new_index
            extracted_parts.append(source_text[old_index:next_index])
        elif start_character is not None and ParserHelper.is_character_at_index(
            source_text, next_index, start_character
        ):
            POGGER.debug("pre-start>>next_index>>$>>", next_index)
            extracted_parts.append(start_character)
            next_index += 1
            nesting_level += 1
        else:
            assert ParserHelper.is_character_at_index(
                source_text, next_index, close_character
            ), "Character at index must be the close character."
            POGGER.debug("pre-close>>next_index>>$>>", next_index)
            extracted_parts.append(close_character)
            next_index += 1
            nesting_level -= 1
        nexter_index, new_data = ParserHelper.collect_until_one_of_characters_verified(
            source_text, next_index, break_characters
        )
        extracted_parts.append(new_data)
        return nexter_index, nesting_level

    # pylint: enable=too-many-arguments

    @staticmethod
    def append_text(
        string_to_append_to: str,
        text_to_append: str,
        alternate_escape_map: Optional[Dict[str, str]] = None,
        add_text_signature: bool = True,
    ) -> str:
        """
        Append the text to the given string, doing any needed encoding as we go.
        """

        if not alternate_escape_map:
            alternate_escape_map = InlineHelper.__html_character_escape_map
        key_map = "".join(alternate_escape_map.keys())
        start_index, text_parts = 0, [string_to_append_to]
        next_index = ParserHelper.index_any_of(text_to_append, key_map, start_index)
        while next_index != -1:
            escaped_part = alternate_escape_map[text_to_append[next_index]]
            text_parts.extend(
                [
                    text_to_append[start_index:next_index],
                    (
                        ParserHelper.create_replacement_markers(
                            text_to_append[next_index], escaped_part
                        )
                        if add_text_signature
                        else escaped_part
                    ),
                ]
            )

            start_index = next_index + 1
            next_index = ParserHelper.index_any_of(text_to_append, key_map, start_index)

        if start_index < len(text_to_append):
            text_parts.append(text_to_append[start_index:])

        return "".join(text_parts)

    # pylint: disable=too-many-arguments
    @staticmethod
    def pdff(
        current_line_source_text: str,
        current_line_first_word: str,
        current_line_leading_space_index: Optional[int],
        adj_tabified_text: str,
        stop_character_in_tabified_index: int,
        current_line_leading_space: Optional[str],
    ) -> int:
        """
        This method needs a better name.
        """
        first_word_index = current_line_source_text.rfind(current_line_first_word)
        POGGER.debug("first_word_index>:$:<", first_word_index)
        first_word_count = 0
        while (
            first_word_index != current_line_leading_space_index
            and first_word_index > 0
        ):
            first_word_index = current_line_source_text.rfind(
                current_line_first_word, 0, first_word_index - 1
            )
            POGGER.debug("first_word_index>:$:<", first_word_index)
            first_word_count += 1
        POGGER.debug("first_word_count>:$:<", first_word_count)
        assert (
            first_word_index == current_line_leading_space_index
        ), "Indices must match up."

        tabified_start_index = adj_tabified_text.rfind(
            current_line_first_word, 0, stop_character_in_tabified_index
        )
        POGGER.debug(
            "tabified_start_index=$,stop=$..>:$:<",
            tabified_start_index,
            stop_character_in_tabified_index,
            adj_tabified_text[tabified_start_index:stop_character_in_tabified_index],
        )
        for _ in range(first_word_count):
            tabified_start_index = adj_tabified_text.rfind(
                current_line_first_word, 0, tabified_start_index - 1
            )
            POGGER.debug(
                "tabified_start_index=$,stop=$..>:$:<",
                tabified_start_index,
                stop_character_in_tabified_index,
                adj_tabified_text[
                    tabified_start_index:stop_character_in_tabified_index
                ],
            )
            assert (
                tabified_start_index <= stop_character_in_tabified_index
            ), "Normal and tabified indices must match up."

        if current_line_leading_space:
            POGGER.debug(
                "adj_tabified_text[$:]>:$:<",
                tabified_start_index,
                adj_tabified_text[tabified_start_index:],
            )
            tabified_start_index, _ = ParserHelper.extract_spaces_from_end(
                adj_tabified_text, tabified_start_index
            )
            POGGER.debug(
                "adj_tabified_text[$:]>:$:<",
                tabified_start_index,
                adj_tabified_text[tabified_start_index:],
            )

        return tabified_start_index

    # pylint: enable=too-many-arguments

    @staticmethod
    def xdf(tabified_text: str, newlines_encountered: int) -> Tuple[str, int]:
        """
        This method needs a better name.
        """
        start_index = 0
        for _ in range(newlines_encountered):
            next_index = tabified_text.find("\n", start_index)
            assert next_index != -1, "Next newline must be found."
            start_index = next_index + 1

        next_index = tabified_text.find("\n", start_index)
        assert next_index == -1, "Next newline must not be found."

        return tabified_text[start_index:], start_index

    @staticmethod
    def xdg(tabified_text: str, newlines_encountered: int) -> Tuple[str, int]:
        """
        This method needs a better name.
        """
        line_start_index = 0
        for _ in range(newlines_encountered):
            line_end_index = tabified_text.find("\n", line_start_index)
            line_start_index = line_end_index + 1
        POGGER.debug("line_start_index>:$:<", line_start_index)
        line_end_index = tabified_text.find("\n", line_start_index)
        return (
            (
                tabified_text[line_start_index:line_end_index]
                if line_end_index != -1
                else tabified_text[line_start_index:]
            ),
            line_start_index,
        )

    @staticmethod
    def calculate_word_index(
        current_line_tabified_text: str,
        source_text_word: str,
        find_word_count: int,
        source_text_spaces: str,
    ) -> int:
        """
        Calculate the location of a given word.
        """
        POGGER.debug("source_text_word>:$:<", source_text_word)
        found_word_index = current_line_tabified_text.find(source_text_word)
        POGGER.debug(
            "[$]-->current_line_source_text[$:]>:$:<",
            0,
            found_word_index,
            current_line_tabified_text[found_word_index:],
        )
        for found_word_count in range(find_word_count):
            found_word_index = current_line_tabified_text.find(
                source_text_word, found_word_index + 1
            )
            POGGER.debug(
                "[$]-->current_line_source_text[$:]>:$:<",
                found_word_count,
                found_word_index,
                current_line_tabified_text[found_word_index:],
            )
            assert found_word_index != -1, "Found index must be set to a valid index."

        if source_text_spaces:
            POGGER.debug(
                "current_line_source_text[$:]>:$:<",
                found_word_index,
                current_line_tabified_text[found_word_index:],
            )
            found_word_index, _ = ParserHelper.extract_spaces_from_end(
                current_line_tabified_text, found_word_index
            )
            POGGER.debug(
                "current_line_source_text[$:]>:$:<",
                found_word_index,
                current_line_tabified_text[found_word_index:],
            )
        return found_word_index



"""
Module to provide classes to deal with extensions.
"""

import argparse
import logging
import re
from typing import Dict, List, Optional, Set, Tuple

from application_properties import ApplicationProperties, ApplicationPropertiesFacade
from columnar import columnar

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.extensions.disallowed_raw_html import MarkdownDisallowRawHtmlExtension
from pymarkdown.extensions.extended_autolinks import MarkdownExtendedAutolinksExtension
from pymarkdown.extensions.extension_one import DebugExtension
from pymarkdown.extensions.front_matter_extension import FrontMatterExtension
from pymarkdown.extensions.markdown_strikethrough import MarkdownStrikeThroughExtension
from pymarkdown.extensions.markdown_tables import MarkdownTablesExtension
from pymarkdown.extensions.pragma_token import PragmaExtension
from pymarkdown.extensions.task_list_items import MarkdownTaskListItemsExtension
from pymarkdown.general.main_presentation import MainPresentation
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade
from pymarkdown.plugin_manager.bad_plugin_error import BadPluginError
from pymarkdown.return_code_helper import ApplicationResult

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class ExtensionManager:
    """
    Manager object to take care of loading and accessing extension modules.
    """

    __extensions_prefix = "extensions"
    __argparse_subparser: Optional[argparse.ArgumentParser] = None
    __root_subparser_name = "em_subcommand"
    __id_regex = re.compile("^[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9]$")

    __DEBUG_EXTENSION = DebugExtension()

    def __init__(
        self,
        presentation: MainPresentation,
    ) -> None:
        self.__presentation = presentation

        self.__extension_objects: Dict[str, ParserExtension] = {}
        self.__extension_details: Dict[str, ExtensionDetails] = {}
        self.__enabled_extensions: List[str] = []
        self.__properties: Optional[ApplicationProperties] = None
        self.__is_front_matter_enabled: bool = False
        self.__is_linter_pragmas_enabled: bool = False
        self.__is_disallow_raw_html_enabled: bool = False
        self.__is_task_list_items_enabled: bool = False
        self.__is_strike_through_enabled: bool = False
        self.__is_extended_autolinks_enabled: bool = False
        self.__is_tables_enabled: bool = False

    def initialize(
        self,
        properties: ApplicationProperties,
    ) -> None:
        """
        Initializes the manager by adding extensions and registering them.
        """

        self.__properties = properties

        all_extensions: List[ParserExtension] = [
            FrontMatterExtension(),
            PragmaExtension(),
            MarkdownTablesExtension(),
            MarkdownTaskListItemsExtension(),
            MarkdownStrikeThroughExtension(),
            MarkdownExtendedAutolinksExtension(),
            MarkdownDisallowRawHtmlExtension(),
            ExtensionManager.__DEBUG_EXTENSION,
        ]

        for next_extension_object in all_extensions:
            next_extension = next_extension_object.get_details()

            assert (
                next_extension.extension_interface_version
                == ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC
            ), "Only the basic extension version is supported."
            self.__extension_details[next_extension.extension_id] = next_extension
            self.__extension_objects[next_extension.extension_id] = (
                next_extension_object
            )
            _ = next_extension.extension_interface_version
            _ = next_extension.extension_configuration

    def apply_configuration(self, enable_extensions_from_command_line: str) -> None:
        """
        Apply any supplied configuration to each of the enabled extensions.
        """
        command_line_enabled_extensions: Set[str] = set()
        if enable_extensions_from_command_line:
            command_line_enabled_extensions.update(
                next_rule_identifier.strip()
                for next_rule_identifier in enable_extensions_from_command_line.lower().split(
                    ","
                )
            )

        if invalid_extension_ids := command_line_enabled_extensions - set(
            self.__extension_details.keys()
        ):
            raise BadPluginError(
                formatted_message=f"Invalid extensions ids supplied to the --enable-extensions command-line option: {','.join(invalid_extension_ids)}."
            )

        for (
            next_extension_id,
            next_extension_detail,
        ) in self.__extension_details.items():
            (
                is_enabled,
                extension_specific_facade,
            ) = self.__determine_if_extension_enabled(
                next_extension_detail, command_line_enabled_extensions
            )
            LOGGER.info("extension %s: enabled=%s", next_extension_id, is_enabled)
            if is_enabled:
                self.__enabled_extensions.append(next_extension_id)

                next_extension_object = self.__extension_objects[next_extension_id]
                next_extension_object.apply_configuration(extension_specific_facade)
        self.__is_front_matter_enabled = (
            FrontMatterExtension().get_identifier() in self.__enabled_extensions
        )
        self.__is_linter_pragmas_enabled = (
            PragmaExtension().get_identifier() in self.__enabled_extensions
        )
        self.__is_disallow_raw_html_enabled = (
            MarkdownDisallowRawHtmlExtension().get_identifier()
            in self.__enabled_extensions
        )
        self.__is_task_list_items_enabled = (
            MarkdownTaskListItemsExtension().get_identifier()
            in self.__enabled_extensions
        )
        self.__is_strike_through_enabled = (
            MarkdownStrikeThroughExtension().get_identifier()
            in self.__enabled_extensions
        )
        self.__is_extended_autolinks_enabled = (
            MarkdownExtendedAutolinksExtension().get_identifier()
            in self.__enabled_extensions
        )
        self.__is_tables_enabled = (
            MarkdownTablesExtension().get_identifier() in self.__enabled_extensions
        )

    def get_extension_instance(self, extension_id: str) -> ParserExtension:
        """
        Get instances of the extension objects.
        """
        return self.__extension_objects[extension_id]

    @property
    def is_front_matter_enabled(self) -> bool:
        """
        Check to see if front-matter support is enabled.
        """
        return self.__is_front_matter_enabled

    @property
    def is_linter_pragmas_enabled(self) -> bool:
        """
        Check to see if linter-pragmas support is enabled.
        """
        return self.__is_linter_pragmas_enabled

    @property
    def is_disallow_raw_html_enabled(self) -> bool:
        """
        Check to see if disallow_raw_html support is enabled.
        """
        return self.__is_disallow_raw_html_enabled

    @property
    def is_task_list_items_enabled(self) -> bool:
        """
        Check to see if task list items support is enabled.
        """
        return self.__is_task_list_items_enabled

    @property
    def is_strike_through_enabled(self) -> bool:
        """
        Check to see if strike through support is enabled.
        """
        return self.__is_strike_through_enabled

    @property
    def is_extended_autolinks_enabled(self) -> bool:
        """
        Check to see if extended autolinks support is enabled.
        """
        return self.__is_extended_autolinks_enabled

    @property
    def is_tables_enabled(self) -> bool:
        """
        Check to see if tables support is enabled.
        """
        return self.__is_tables_enabled

    @staticmethod
    def argparse_subparser_name() -> str:
        """
        Gets the name of the subparser used to handle these extensions.
        """
        return "extensions"

    def handle_argparse_subparser(self, args: argparse.Namespace) -> ApplicationResult:
        """
        Handle the parsing for this subparser.
        """
        subparser_value = getattr(args, ExtensionManager.__root_subparser_name)
        if subparser_value == "list":
            return self.__handle_argparse_subparser_list(args)
        if subparser_value == "info":
            return self.__handle_argparse_subparser_info(args)
        assert (
            ExtensionManager.__argparse_subparser is not None
        ), "Subparser must be defined by this point."
        ExtensionManager.__argparse_subparser.print_help()
        return ApplicationResult.COMMAND_LINE_ERROR

    def __handle_argparse_subparser_list(
        self, args: argparse.Namespace
    ) -> ApplicationResult:
        list_re = None
        if args.list_filter:
            list_re = re.compile(
                "^" + args.list_filter.replace("*", ".*").replace("?", ".") + "$"
            )

        names, show_rows = list(self.__extension_details.keys()), []
        names.sort()
        for next_extension_name in names:
            next_extension = self.__extension_details[next_extension_name]

            if (
                next_extension.extension_id
                == ExtensionManager.__DEBUG_EXTENSION.get_identifier()
                or (
                    next_extension.extension_version
                    == ExtensionManagerConstants.EXTENSION_VERSION_NOT_IMPLEMENTED
                    and not args.show_all
                )
            ):
                continue

            if list_re.match(next_extension_name) if list_re else True:
                is_enabled_now = next_extension_name in self.__enabled_extensions
                display_row = [
                    next_extension.extension_id,
                    next_extension.extension_name,
                    str(next_extension.extension_enabled_by_default),
                    str(is_enabled_now),
                    next_extension.extension_version,
                ]
                show_rows.append(display_row)

        if show_rows:
            headers = [
                "id",
                "name",
                "enabled\n(default)",
                "enabled\n(current)",
                "version",
            ]
            self.__print_column_output(headers, show_rows)
            return ApplicationResult.SUCCESS
        self.__presentation.print_system_error(
            f"No extension identifier matches the pattern '{args.list_filter}'."
        )
        return ApplicationResult.NO_FILES_TO_SCAN

    def __handle_argparse_subparser_info(
        self, args: argparse.Namespace
    ) -> ApplicationResult:
        if args.info_filter not in self.__extension_details:
            self.__presentation.print_system_error(
                f"Unable to find an extension with an id of '{args.info_filter}'."
            )
            return ApplicationResult.NO_FILES_TO_SCAN

        found_extension = self.__extension_details[args.info_filter]
        show_rows: List[List[str]] = [
            ["Id", found_extension.extension_id],
            ["Name", found_extension.extension_name],
            ["Short Description", found_extension.extension_description],
            ["Description Url", str(found_extension.extension_url)],
        ]

        headers = ["Item", "Description"]
        self.__print_column_output(headers, show_rows)
        return ApplicationResult.SUCCESS

    def __print_column_output(
        self, headers: List[str], show_rows: List[List[str]]
    ) -> None:
        table = columnar(show_rows, headers, no_borders=True)
        split_rows = table.split(ParserHelper.newline_character)
        new_rows = [next_row.rstrip() for next_row in split_rows]
        self.__presentation.print_system_output(
            ParserHelper.newline_character.join(new_rows)
        )

    @staticmethod
    def __list_filter_type(argument: str) -> str:
        test_argument = argument.replace("*", "").replace("?", "")
        if ExtensionManager.__id_regex.match(test_argument):
            return argument
        raise argparse.ArgumentTypeError(
            f"Value '{argument}' is not a valid pattern for an id."
        )

    @staticmethod
    def __info_filter_type(argument: str) -> str:
        if ExtensionManager.__id_regex.match(argument):
            return argument
        raise argparse.ArgumentTypeError(f"Value '{argument}' is not a valid id.")

    @staticmethod
    def add_argparse_subparser(subparsers: argparse._SubParsersAction) -> None:  # type: ignore
        """
        Populate the argparse tree to allow for plugin support.
        """
        new_sub_parser = subparsers.add_parser(
            ExtensionManager.argparse_subparser_name(), help="extension commands"
        )
        ExtensionManager.__argparse_subparser = new_sub_parser
        plugin_subparsers = new_sub_parser.add_subparsers(
            dest=ExtensionManager.__root_subparser_name
        )

        sub_sub_parser = plugin_subparsers.add_parser(
            "list", help="list the available extensions"
        )
        sub_sub_parser.add_argument(
            "--all",
            dest="show_all",
            action="store_true",
            default=False,
            help="show all loaded extensions (default is False)",
        )
        sub_sub_parser.add_argument(
            dest="list_filter",
            default=None,
            help="filter",
            nargs="?",
            type=ExtensionManager.__list_filter_type,
        )
        sub_sub_parser = plugin_subparsers.add_parser(
            "info", help="information on a specific extension"
        )
        sub_sub_parser.add_argument(
            dest="info_filter",
            default=None,
            type=ExtensionManager.__info_filter_type,
            help="an id",
        )

    def __determine_if_extension_enabled(
        self,
        extension_object: ExtensionDetails,
        command_line_enabled_extensions: Set[str],
    ) -> Tuple[bool, MyApplicationPropertiesFacade]:
        """
        Given the enable and disable rule values, evaluate the enabled or disabled
        state of the extension in proper order.
        """

        assert (
            self.__properties is not None
        ), "Properties must be defined by this point."
        plugin_section_title = (
            f"{ExtensionManager.__extensions_prefix}{self.__properties.separator}"
            + f"{extension_object.extension_id}{self.__properties.separator}"
        )
        extension_specific_facade = MyApplicationPropertiesFacade(
            ApplicationPropertiesFacade(self.__properties, plugin_section_title)
        )

        LOGGER.debug(
            "Extension '%s'",
            extension_object.extension_id,
        )
        new_value = None
        if extension_object.extension_id in command_line_enabled_extensions:
            new_value = True
        else:
            new_value = extension_specific_facade.get_boolean_property(
                "enabled", default_value=None
            )
            if new_value is None:
                LOGGER.debug(
                    "No other enable state found, setting to default of '%s'.",
                    str(extension_object.extension_enabled_by_default),
                )
            else:
                LOGGER.debug(
                    "Extension specific key 'enabled' found, value is '%s'.",
                    str(new_value),
                )

        return (
            extension_object.extension_enabled_by_default
            if new_value is None
            else new_value
        ), extension_specific_facade


# pylint: enable=too-many-instance-attributes



"""
Module to provide for an encapsulation of the atx heading element.
"""

from typing import Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class AtxHeadingMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the atx heading element.
    """

    def __init__(
        self,
        hash_count: int,
        remove_trailing_count: int,
        extracted_whitespace: str,
        position_marker: PositionMarker,
    ) -> None:
        self.__hash_count, self.__remove_trailing_count = (
            hash_count,
            remove_trailing_count,
        )

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_atx_heading,
            "",
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
            can_force_close=False,
        )
        self.__compose_extra_data_field()

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_atx_heading

    # pylint: enable=protected-access

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if (
            field_name == "hash_count"
            and isinstance(field_value, int)
            and 1 <= field_value <= 6
        ):
            self.__hash_count = field_value
            self.__compose_extra_data_field()
            return True

        # Handle extracted_whitespace
        return super()._modify_token(field_name, field_value)

    @property
    def hash_count(self) -> int:
        """
        Returns the number of hash marks specified at the start of the line.
        """
        return self.__hash_count

    @property
    def remove_trailing_count(self) -> int:
        """
        Returns the number of hash marks specified at the end of the line.
        """
        return self.__remove_trailing_count

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [
                    str(self.__hash_count),
                    str(self.__remove_trailing_count),
                    self.extracted_whitespace,
                ]
            )
        )

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            AtxHeadingMarkdownToken,
            AtxHeadingMarkdownToken.__rehydrate_atx_heading,
            AtxHeadingMarkdownToken.__rehydrate_atx_heading_end,
        )

    @staticmethod
    def __rehydrate_atx_heading(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the atx heading block from the token.
        """
        _ = previous_token

        context.block_stack.append(current_token)
        current_start_token = cast(AtxHeadingMarkdownToken, current_token)
        return f'{current_start_token.extracted_whitespace}{ParserHelper.repeat_string("#", current_start_token.hash_count)}'

    @staticmethod
    def __rehydrate_atx_heading_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the end of the atx heading block from the token.
        """
        _ = (previous_token, next_token)

        current_end_token = cast(EndMarkdownToken, current_token)
        current_start_token = cast(
            AtxHeadingMarkdownToken, current_end_token.start_markdown_token
        )

        del context.block_stack[-1]
        assert (
            current_end_token.extra_end_data is not None
        ), "extra_end_data must be defined by now."
        return "".join(
            [
                current_end_token.extra_end_data,
                (
                    ParserHelper.repeat_string(
                        "#", current_start_token.remove_trailing_count
                    )
                    if current_start_token.remove_trailing_count
                    else ""
                ),
                current_end_token.extracted_whitespace,
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            AtxHeadingMarkdownToken,
            AtxHeadingMarkdownToken.__handle_start_atx_heading_token,
            AtxHeadingMarkdownToken.__handle_end_atx_heading_token,
        )

    @staticmethod
    def __handle_start_atx_heading_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        atx_token = cast(AtxHeadingMarkdownToken, next_token)
        previous_token = transform_state.actual_tokens[
            transform_state.actual_token_index - 1
        ]

        token_parts = [output_html]
        if (output_html.endswith("</ol>") or output_html.endswith("</ul>")) or (
            previous_token.is_paragraph_end and not transform_state.is_in_loose_list
        ):
            token_parts.append(ParserHelper.newline_character)
        token_parts.extend(["<h", str(atx_token.hash_count), ">"])
        return "".join(token_parts)

    @staticmethod
    def __handle_end_atx_heading_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        fenced_token_index = transform_state.actual_token_index - 1
        while not transform_state.actual_tokens[fenced_token_index].is_atx_heading:
            fenced_token_index -= 1
        fenced_token = cast(
            SetextHeadingMarkdownToken,
            transform_state.actual_tokens[fenced_token_index],
        )

        return "".join(
            [
                output_html,
                "</h",
                str(fenced_token.hash_count),
                ">",
                ParserHelper.newline_character,
            ]
        )
